/*
 * Copyright (c) 1985 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

/*
 * Grammar for FTP commands.
 * See RFC 765.
 **********************************************************************
 * HISTORY
 * $Log:	ftpcmd.y,v $
 * Revision 1.1  92/04/09  01:31:29  vac
 * Initial revision
 * 
 * Revision 2.5  91/05/08  12:53:34  dimitris
 * 	Modified to make sure that the anonymous password sent 
 * 	has a '@' in it before it is logged. If it doesn't the
 * 	anonymous login is rejected.
 * 	[91/05/08  12:52:59  dimitris]
 * 
 * Revision 2.4  91/04/11  16:38:11  dimitris
 * 	Modified to record the password for anonymous ftp
 * 	(since this is usually the username).
 * 	[91/04/11  16:37:47  dimitris]
 * 
 * Revision 2.3  88/12/07  17:29:01  dimitris
 * 	Modified to look for "anon" as opposed to "ftpguest"
 * 	before allowing guest login.
 * 	[88/12/07  17:24:45  dimitris]
 * 
 * 09-May-86  Rudy Nedved (ern) at Carnegie-Mellon University
 *	Updated for latest ftp RFC which is RFC 959.
 *
 **********************************************************************
 */

%{

#ifndef lint
static	char sccsid[] = "@(#)ftpcmd.y	5.7 (Berkeley) 5/28/86";
#endif

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#include <arpa/ftp.h>

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <pwd.h>
#include <setjmp.h>
#include <syslog.h>

extern	struct sockaddr_in data_dest;
extern	int logged_in;
extern	struct passwd *pw;
extern	int guest;
extern	int logging;
extern	int type;
extern	int form;
extern	int debug;
extern	int timeout;
extern  int pdata;
extern	char hostname[];
extern	char *globerr;
extern	int usedefault;
extern	int unique;
extern  int transflag;
extern  char tmpline[];
char	**glob();

static	int cmd_type;
static	int cmd_form;
static	int cmd_bytesz;
char cbuf[512];

char	*index();
%}

%token
	A	B	C	E	F	I
	L	N	P	R	S	T

	SP	CRLF	COMMA	STRING	NUMBER

	USER	PASS	ACCT	REIN	QUIT	PORT
	PASV	TYPE	STRU	MODE	RETR	STOR
	APPE	MLFL	MAIL	MSND	MSOM	MSAM
	MRSQ	MRCP	ALLO	REST	RNFR	RNTO
	ABOR	DELE	CWD	LIST	NLST	SITE
	STAT	HELP	NOOP	CDUP	SMNT	STOU
	RMD	MKD	PWD	SYST

	LEXERR

%start	cmd_list

%%

cmd_list:	/* empty */
	|	cmd_list cmd
	;

cmd:		USER SP username CRLF
		= {
#ifdef	CMUCS
                        if (strcmp((char *) $3, "ftp") ==0 ||
                            strcmp((char *) $3,"anonymous")==0) {
                             if ((pw =getpwnam("anon")) != NULL) {
                                   guest = 1;
                                   reply (331, 
                                   "Guest login ok, send username@node as password.");
                             } else {
                               reply(530, "No anonymous FTP access allowed");
                             }
                        }
        		user((char *) $3);
#else
			extern struct passwd *getpwnam();

			logged_in = 0;
			if (strcmp((char *) $3, "ftp") == 0 ||
			  strcmp((char *) $3, "anonymous") == 0) {
				if ((pw = getpwnam("ftp")) != NULL) {
					guest = 1;
					reply(331,
				  "Guest login ok, send ident as password.");
				}
				else {
					reply(530, "User %s unknown.", $3);
				}
			} else if (checkuser((char *) $3)) {
				guest = 0;
				pw = getpwnam((char *) $3);
				if (pw == NULL) {
					reply(530, "User %s unknown.", $3);
				}
				else {
				    reply(331, "Password required for %s.", $3);
				}
			} else {
				reply(530, "User %s access denied.", $3);
			}
#endif
			free((char *) $3);
		}
	|	PASS SP password CRLF
		= {
			pass((char *) $3);
			free((char *) $3);
		}
        |       PASS CRLF
                = {
                        pass("");
                }
	|	PORT SP host_port CRLF
		= {
			usedefault = 0;
			if (pdata > 0) {
				(void) close(pdata);
			}
			pdata = -1;
			reply(200, "PORT command successful.");
		}
	|	PASV CRLF
		= {
			passive();
		}
	|	TYPE SP type_code CRLF
		= {
			switch (cmd_type) {

			case TYPE_A:
				if (cmd_form == FORM_N) {
#ifdef	CMUCS
					reply(200, "Type set to A N.");
#else
					reply(200, "Type set to A.");
#endif
					type = cmd_type;
					form = cmd_form;
				} else
					reply(504, "Form must be N.");
				break;

			case TYPE_E:
				reply(504, "Type E not implemented.");
				break;

			case TYPE_I:
				reply(200, "Type set to I.");
				type = cmd_type;
				break;

			case TYPE_L:
				if (cmd_bytesz == 8) {
					reply(200,
#ifdef	CMUCS
					    "Type set to L 8.");
#else
					    "Type set to L (byte size 8).");
#endif
					type = cmd_type;
				} else
					reply(504, "Byte size must be 8.");
			}
		}
	|	STRU SP struct_code CRLF
		= {
			switch ($3) {

			case STRU_F:
				reply(200, "STRU F ok.");
				break;

			default:
				reply(504, "Unimplemented STRU type.");
			}
		}
	|	MODE SP mode_code CRLF
		= {
			switch ($3) {

			case MODE_S:
				reply(200, "MODE S ok.");
				break;

			default:
				reply(502, "Unimplemented MODE type.");
			}
		}
	|	ALLO SP NUMBER CRLF
		= {
			reply(202, "ALLO command ignored.");
		}
	|	RETR check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
				retrieve((char *) 0, (char *) $4);
			if ($4 != NULL)
				free((char *) $4);
		}
	|	STOR check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
				store((char *) $4, "w");
			if ($4 != NULL)
				free((char *) $4);
		}
	|	APPE check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
				store((char *) $4, "a");
			if ($4 != NULL)
				free((char *) $4);
		}
	|	NLST check_login CRLF
		= {
			if ($2)
#ifdef	CMUCS
				retrieve("ls", "");   
#else
				retrieve("/bin/ls", "");
#endif
		}
	|	NLST check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
#ifdef	CMUCS
				retrieve("ls %s", (char *) $4);
#else
				retrieve("/bin/ls %s", (char *) $4);
#endif
			if ($4 != NULL)
				free((char *) $4);
		}
	|	LIST check_login CRLF
		= {
			if ($2)
#ifdef	CMUCS
				retrieve("ls -lgL", "");   /* vac  see note 2 lines down */
				/* retrieve("ls -lg", "");    vac */ 
/*  It turns out that /usr/cs/bin/ls has exactly the opposite meaning for the -L flag.
For the CMU FTP this usually good, because users can go to directories in places 
where they can not go to intermediate directories.  Not showing the intermediate 
directories can keep some users from getting confused.  For Alex this is really bad.
for performance.  Do  a "man alexinalex" for more info.
   - vac
*/

#else
				retrieve("/bin/ls -lg", "");
#endif
		}
	|	LIST check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
#ifdef	CMUCS
				retrieve("ls -lgL %s", (char *) $4);  /* vac */
				/*retrieve("ls -lg %s", (char *) $4);    vac */
#else
				retrieve("/bin/ls -lg %s", (char *) $4);
#endif
			if ($4 != NULL)
				free((char *) $4);
		}
	|	DELE check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
				delete((char *) $4);
			if ($4 != NULL)
				free((char *) $4);
		}
	|	ABOR CRLF
		= {
			reply(225, "ABOR command successful.");
		}
	|	CWD check_login CRLF
		= {
			if ($2)
				cwd(pw->pw_dir);
		}
	|	CWD check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
				cwd((char *) $4);
			if ($4 != NULL)
				free((char *) $4);
		}
	|	rename_cmd
	|	HELP CRLF
		= {
			help((char *) 0);
		}
	|	HELP SP STRING CRLF
		= {
			help((char *) $3);
		}
	|	NOOP CRLF
		= {
			reply(200, "NOOP command successful.");
		}
	|	MKD check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
				makedir((char *) $4);
			if ($4 != NULL)
				free((char *) $4);
		}
	|	RMD check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL)
				removedir((char *) $4);
			if ($4 != NULL)
				free((char *) $4);
		}
	|	PWD check_login CRLF
		= {
			if ($2)
				pwd();
		}
	|	CDUP check_login CRLF
		= {
			if ($2)
				cwd("..");
		}
	|	STOU check_login SP pathname CRLF
		= {
			if ($2 && $4 != NULL) {
				unique++;
				store((char *) $4, "w");
				unique = 0;
			}
			if ($4 != NULL)
				free((char *) $4);
		}
	|	QUIT CRLF
		= {
			reply(221, "Goodbye.");
#ifndef VACFASTFTPD
			dologout(0);
#endif
		}
	|	SYST CRLF
		= {
			reply(215, "UNIX");
		}
	|	error CRLF
		= {
			yyerrok;
		}
	;

username:	STRING
	;

password:	STRING
	;

byte_size:	NUMBER
	;

host_port:	NUMBER COMMA NUMBER COMMA NUMBER COMMA NUMBER COMMA 
		NUMBER COMMA NUMBER
		= {
			register char *a, *p;

			a = (char *)&data_dest.sin_addr;
			a[0] = $1; a[1] = $3; a[2] = $5; a[3] = $7;
			p = (char *)&data_dest.sin_port;
			p[0] = $9; p[1] = $11;
			data_dest.sin_family = AF_INET;
		}
	;

form_code:	N
	= {
		$$ = FORM_N;
	}
	|	T
	= {
		$$ = FORM_T;
	}
	|	C
	= {
		$$ = FORM_C;
	}
	;

type_code:	A
	= {
		cmd_type = TYPE_A;
		cmd_form = FORM_N;
	}
	|	A SP form_code
	= {
		cmd_type = TYPE_A;
		cmd_form = $3;
	}
	|	E
	= {
		cmd_type = TYPE_E;
		cmd_form = FORM_N;
	}
	|	E SP form_code
	= {
		cmd_type = TYPE_E;
		cmd_form = $3;
	}
	|	I
	= {
		cmd_type = TYPE_I;
	}
	|	L
	= {
		cmd_type = TYPE_L;
		cmd_bytesz = 8;
	}
	|	L SP byte_size
	= {
		cmd_type = TYPE_L;
		cmd_bytesz = $3;
	}
	/* this is for a bug in the BBN ftp */
	|	L byte_size
	= {
		cmd_type = TYPE_L;
		cmd_bytesz = $2;
	}
	;

struct_code:	F
	= {
		$$ = STRU_F;
	}
	|	R
	= {
		$$ = STRU_R;
	}
	|	P
	= {
		$$ = STRU_P;
	}
	;

mode_code:	S
	= {
		$$ = MODE_S;
	}
	|	B
	= {
		$$ = MODE_B;
	}
	|	C
	= {
		$$ = MODE_C;
	}
	;

pathname:	pathstring
	= {
		/*
		 * Problem: this production is used for all pathname
		 * processing, but only gives a 550 error reply.
		 * This is a valid reply in some cases but not in others.
		 */
		if ($1 && strncmp((char *) $1, "~", 1) == 0) {
			$$ = (int)*glob((char *) $1);
			if (globerr != NULL) {
				reply(550, globerr);
				$$ = NULL;
			}
			free((char *) $1);
		} else
			$$ = $1;
	}
	;

pathstring:	STRING
	;

rename_cmd:	rename_from rename_to
	= {
		if ($1 && $2)
			renamecmd((char *) $1, (char *) $2);
		else
			reply(503, "Bad sequence of commands.");
		if ($1)
			free((char *) $1);
		if ($2)
			free((char *) $2);
	}
	;

rename_from:	RNFR check_login SP pathname CRLF
	= {
		char *from = 0, *renamefrom();

		if ($2 && $4)
			from = renamefrom((char *) $4);
		if (from == 0 && $4)
			free((char *) $4);
		$$ = (int)from;
	}
	;

rename_to:	RNTO SP pathname CRLF
	= {
		$$ = $3;
	}
	;

check_login:	/* empty */
	= {
		if (logged_in)
			$$ = 1;
#ifdef	CMUCS
		else if (oklocal()) {
			freelogin();
			$$ = logged_in;
		}
#endif
		else {
			reply(530, "Please login with USER and PASS.");
			$$ = 0;
		}
	}
	;

%%

extern jmp_buf errcatch;

#define	CMD	0	/* beginning of command */
#define	ARGS	1	/* expect miscellaneous arguments */
#define	STR1	2	/* expect SP followed by STRING */
#define	STR2	3	/* expect STRING */
#define	OSTR	4	/* optional STRING */

struct tab {
	char	*name;
	short	token;
	short	state;
	short	implemented;	/* 1 if command is implemented */
	char	*help;
};

#ifdef	CMUCS
struct tab cmdtab[] = {		/* In order defined in RFC 959 */
	{ "USER", USER, STR1, 1,	"<sp> username" },
	{ "PASS", PASS, OSTR, 1,	"<sp> password" },
	{ "ACCT", ACCT, STR1, 0,	"<sp> account-information" },
	{ "CWD",  CWD,	OSTR, 1,	"[ <sp> pathname ]" },
	{ "XCWD", CWD,	OSTR, 1,	"[ <sp> pathname ]" },
	{ "CDUP", CDUP, ARGS, 1,	"(change to parent directory)" },
	{ "XCUP", CDUP, ARGS, 1,	"(change to parent directory)" },
	{ "SMNT", SMNT, STR1, 0,	"<sp> pathname" },
	{ "QUIT", QUIT, ARGS, 1,	"(terminate service)" },
	{ "REIN", REIN, ARGS, 0,	"(reinitialize server state)" },
	{ "PORT", PORT, ARGS, 1,	"<sp> host-port" },
	{ "PASV", PASV, ARGS, 0,	"(set server in passive mode)" },
	{ "TYPE", TYPE, ARGS, 1,	"<sp> type-code" },
	{ "STRU", STRU, ARGS, 1,	"<sp> structure-code" },
	{ "MODE", MODE, ARGS, 1,	"<sp> mode-code" },
	{ "RETR", RETR, STR1, 1,	"<sp> pathname" },
	{ "STOR", STOR, STR1, 1,	"<sp> pathname" },
	{ "STOU", STOU, STR1, 0,	"(store unique file)" },
	{ "APPE", APPE, STR1, 1,	"<sp> pathname" },
	{ "ALLO", ALLO, ARGS, 1,	"<sp> file-size" },
	{ "REST", REST, STR1, 0,	"<sp> marker" },
	{ "RNFR", RNFR, STR1, 1,	"<sp> pathname" },
	{ "RNTO", RNTO, STR1, 1,	"<sp> pathname" },
	{ "ABOR", ABOR, ARGS, 0,	"(abort operation)" },
	{ "DELE", DELE, STR1, 1,	"<sp> pathname" },
	{ "RMD",  RMD,	STR1, 1,	"<sp> pathname" },
	{ "XRMD", RMD,	STR1, 1,	"<sp> pathname" },
	{ "MKD",  MKD,	STR1, 1,	"<sp> pathname" },
	{ "XMKD", MKD,	STR1, 1,	"<sp> pathname" },
	{ "PWD",  PWD,	ARGS, 1,	"(print current working directory)" },
	{ "XPWD", PWD,	ARGS, 1,	"(print current working directory)" },
	{ "LIST", LIST, OSTR, 1,	"[ <sp> pathname ]" },
	{ "NLST", NLST, OSTR, 1,	"[ <sp> pathname ]" },
	{ "SITE", SITE, STR1, 0,	"<sp> <string>" },
	{ "SYST", SYST, ARGS, 0,	"(get operating system name)" },
	{ "STAT", STAT, OSTR, 0,	"[ <sp> <pathname> ]" },
	{ "HELP", HELP, OSTR, 1,	"[ <sp> <string> ]" },
	{ "NOOP", NOOP, ARGS, 1,	"(no operation)" },
	{ NULL,   0,    0,    0,	0 }
};
#else
struct tab cmdtab[] = {		/* In order defined in RFC 765 */
	{ "USER", USER, STR1, 1,	"<sp> username" },
	{ "PASS", PASS, STR1, 1,	"<sp> password" },
	{ "ACCT", ACCT, STR1, 0,	"(specify account)" },
	{ "REIN", REIN, ARGS, 0,	"(reinitialize server state)" },
	{ "QUIT", QUIT, ARGS, 1,	"(terminate service)", },
	{ "PORT", PORT, ARGS, 1,	"<sp> b0, b1, b2, b3, b4" },
	{ "PASV", PASV, ARGS, 1,	"(set server in passive mode)" },
	{ "TYPE", TYPE, ARGS, 1,	"<sp> [ A | E | I | L ]" },
	{ "STRU", STRU, ARGS, 1,	"(specify file structure)" },
	{ "MODE", MODE, ARGS, 1,	"(specify transfer mode)" },
	{ "RETR", RETR, STR1, 1,	"<sp> file-name" },
	{ "STOR", STOR, STR1, 1,	"<sp> file-name" },
	{ "APPE", APPE, STR1, 1,	"<sp> file-name" },
	{ "MLFL", MLFL, OSTR, 0,	"(mail file)" },
	{ "MAIL", MAIL, OSTR, 0,	"(mail to user)" },
	{ "MSND", MSND, OSTR, 0,	"(mail send to terminal)" },
	{ "MSOM", MSOM, OSTR, 0,	"(mail send to terminal or mailbox)" },
	{ "MSAM", MSAM, OSTR, 0,	"(mail send to terminal and mailbox)" },
	{ "MRSQ", MRSQ, OSTR, 0,	"(mail recipient scheme question)" },
	{ "MRCP", MRCP, STR1, 0,	"(mail recipient)" },
	{ "ALLO", ALLO, ARGS, 1,	"allocate storage (vacuously)" },
	{ "REST", REST, STR1, 0,	"(restart command)" },
	{ "RNFR", RNFR, STR1, 1,	"<sp> file-name" },
	{ "RNTO", RNTO, STR1, 1,	"<sp> file-name" },
	{ "ABOR", ABOR, ARGS, 1,	"(abort operation)" },
	{ "DELE", DELE, STR1, 1,	"<sp> file-name" },
	{ "CWD",  CWD,  OSTR, 1,	"[ <sp> directory-name]" },
	{ "XCWD", CWD,	OSTR, 1,	"[ <sp> directory-name ]" },
	{ "LIST", LIST, OSTR, 1,	"[ <sp> path-name ]" },
	{ "NLST", NLST, OSTR, 1,	"[ <sp> path-name ]" },
	{ "SITE", SITE, STR1, 0,	"(get site parameters)" },
	{ "STAT", STAT, OSTR, 0,	"(get server status)" },
	{ "HELP", HELP, OSTR, 1,	"[ <sp> <string> ]" },
	{ "NOOP", NOOP, ARGS, 1,	"" },
	{ "MKD",  XMKD, STR1, 1,	"<sp> path-name" },
	{ "XMKD", XMKD, STR1, 1,	"<sp> path-name" },
	{ "RMD",  XRMD, STR1, 1,	"<sp> path-name" },
	{ "XRMD", XRMD, STR1, 1,	"<sp> path-name" },
	{ "PWD",  XPWD, ARGS, 1,	"(return current directory)" },
	{ "XPWD", XPWD, ARGS, 1,	"(return current directory)" },
	{ "CDUP", XCUP, ARGS, 1,	"(change to parent directory)" },
	{ "XCUP", XCUP, ARGS, 1,	"(change to parent directory)" },
	{ "STOU", STOU, STR1, 1,	"<sp> file-name" },
	{ NULL,   0,    0,    0,	0 }
};
#endif

struct tab *
lookup(cmd)
	char *cmd;
{
	register struct tab *p;

	for (p = cmdtab; p->name != NULL; p++)
		if (strcmp(cmd, p->name) == 0)
			return (p);
	return (0);
}

#include <arpa/telnet.h>

/*
 * getline - a hacked up version of fgets to ignore TELNET escape codes.
 */
char *
getline(s, n, iop)
	char *s;
	register FILE *iop;
{
	register c;
	register char *cs;

	cs = s;
/* tmpline may contain saved command from urgent mode interruption */
	for (c = 0; tmpline[c] != '\0' && --n > 0; ++c) {
		*cs++ = tmpline[c];
		if (tmpline[c] == '\n') {
			*cs++ = '\0';
			if (debug) {
				syslog(LOG_DEBUG, "FTPD: command: %s", s);
			}
			tmpline[0] = '\0';
			return(s);
		}
		if (c == 0) {
			tmpline[0] = '\0';
		}
	}
	while (--n > 0 && (c = getc(iop)) != EOF) {
		c = 0377 & c;
		while (c == IAC) {
			switch (c = 0377 & getc(iop)) {
			case WILL:
			case WONT:
				c = 0377 & getc(iop);
				printf("%c%c%c", IAC, WONT, c);
				(void) fflush(stdout);
				break;
			case DO:
			case DONT:
				c = 0377 & getc(iop);
				printf("%c%c%c", IAC, DONT, c);
				(void) fflush(stdout);
				break;
			default:
				break;
			}
			c = 0377 & getc(iop); /* try next character */
		}
#ifdef	CMUCS
		if (c == '\n' && logging) {
			*cs = '\0';
			if (strncasecmp(s, "pass", 4) == 0){
				if (!guest) 
				   syslog(LOG_INFO, "%4.4s XXXXXX\n", s);
				else
				   if (!index(s,'@')) {
					syslog(LOG_INFO, "Bad address format");
		                        reply(530, "Bad address format");
#ifndef VACFASTFTPD
		                        dologout(0);
#endif
				  } else
					syslog(LOG_INFO, "%s\n", s);
			}
			else
				syslog(LOG_INFO, "%s\n", s);
		}
#endif
		*cs++ = c;
		if (c=='\n')
			break;
	}
	if (c == EOF && cs == s)
		return (NULL);
	*cs++ = '\0';
	if (debug) {
		syslog(LOG_DEBUG, "FTPD: command: %s", s);
	}
	return (s);
}

static int
toolong()
{
	time_t now;
	extern char *ctime();
	extern time_t time();

	reply(421,
	  "Timeout (%d seconds): closing control connection.", timeout);
	(void) time(&now);
	if (logging) {
		syslog(LOG_INFO,
			"FTPD: User %s timed out after %d seconds at %s",
			(pw ? pw -> pw_name : "unknown"), timeout, ctime(&now));
	}
#ifndef VACFASTFTPD
	dologout(1);
#endif
}

yylex()
{
	static int cpos, state;
	register char *cp;
	register struct tab *p;
	int n;
	char c;

	for (;;) {
#ifdef	CMUCS
		if (state == OSTR || state == STR1 ||
		    state == STR2 || state == ARGS) {
			/* ignore random white space */
			while (cbuf[cpos] == ' ' && 
			       (cbuf[cpos+1] == ' ' || cbuf[cpos+1] == '\n'))
				cpos++;
		}
#endif
		switch (state) {

		case CMD:
			(void) signal(SIGALRM, toolong);
			(void) alarm((unsigned) timeout);
			if (getline(cbuf, sizeof(cbuf)-1, stdin) == NULL) {
				reply(221, "You could at least say goodbye.");
#ifndef VACFASTFTPD
				dologout(0);
#endif
			}
			(void) alarm(0);
#ifdef	CMUCS
			if (cp = index(cbuf, '\r')) {
				*cp++ = '\n';
				*cp = 0;
			}
			/* find end of command token */
			cp = index(cbuf, ' ');
			if (cp == 0)
			    cp = index(cbuf, '\n');
			cpos = cp - cbuf;
#else
			if (index(cbuf, '\r')) {
				cp = index(cbuf, '\r');
				cp[0] = '\n'; cp[1] = 0;
			}
			if (index(cbuf, ' '))
				cpos = index(cbuf, ' ') - cbuf;
			else
				cpos = index(cbuf, '\n') - cbuf;
			if (cpos == 0) {
				cpos = 4;
			}
#endif
			c = cbuf[cpos];
			cbuf[cpos] = '\0';
			upper(cbuf);
			p = lookup(cbuf);
			cbuf[cpos] = c;
#ifdef	CMUCS
			/* ignore random whitespace after command token */
			while (cbuf[cpos] == ' ' && 
			       (cbuf[cpos+1] == ' ' || cbuf[cpos+1] == '\n'))
				cpos++;
			/* process the command token */
#endif
			if (p != 0) {
				if (p->implemented == 0) {
					nack(p->name);
					longjmp(errcatch,0);
					/* NOTREACHED */
				}
				state = p->state;
				yylval = (int) p->name;
				return (p->token);
			}
			break;

		case OSTR:
			if (cbuf[cpos] == '\n') {
				state = CMD;
				return (CRLF);
			}
#ifdef	CMUCS
			/* ignore random white space */
			while (cbuf[cpos] == ' ' && 
			       (cbuf[cpos+1] == ' ' || cbuf[cpos+1] == '\n'))
				cpos++;
#endif
			/* FALL THRU */

		case STR1:
			if (cbuf[cpos] == ' ') {
#ifdef	CMUCS
				while (cbuf[cpos] == ' ')
#endif
				cpos++;
				state = STR2;
				return (SP);
			}
			break;

		case STR2:
			cp = &cbuf[cpos];
			n = strlen(cp);
			cpos += n - 1;
			/*
			 * Make sure the string is nonempty and \n terminated.
			 */
			if (n > 1 && cbuf[cpos] == '\n') {
				cbuf[cpos] = '\0';
				yylval = copy(cp);
				cbuf[cpos] = '\n';
				state = ARGS;
				return (STRING);
			}
			break;

		case ARGS:
			if (isdigit(cbuf[cpos])) {
				cp = &cbuf[cpos];
				while (isdigit(cbuf[++cpos]))
					;
				c = cbuf[cpos];
				cbuf[cpos] = '\0';
				yylval = atoi(cp);
				cbuf[cpos] = c;
				return (NUMBER);
			}
			switch (cbuf[cpos++]) {

			case '\n':
				state = CMD;
				return (CRLF);

			case ' ':
				return (SP);

			case ',':
				return (COMMA);

			case 'A':
			case 'a':
				return (A);

			case 'B':
			case 'b':
				return (B);

			case 'C':
			case 'c':
				return (C);

			case 'E':
			case 'e':
				return (E);

			case 'F':
			case 'f':
				return (F);

			case 'I':
			case 'i':
				return (I);

			case 'L':
			case 'l':
				return (L);

			case 'N':
			case 'n':
				return (N);

			case 'P':
			case 'p':
				return (P);

			case 'R':
			case 'r':
				return (R);

			case 'S':
			case 's':
				return (S);

			case 'T':
			case 't':
				return (T);

			}
			break;

		default:
			fatal("Unknown state in scanner.");
		}
		yyerror((char *) 0);
		state = CMD;
		longjmp(errcatch,0);
	}
}

upper(s)
	char *s;
{
	while (*s != '\0') {
		if (islower(*s))
			*s = toupper(*s);
		s++;
	}
}

copy(s)
	char *s;
{
	char *p;
	extern char *malloc(), *strcpy();

	p = malloc((unsigned) strlen(s) + 1);
	if (p == NULL)
		fatal("Ran out of memory.");
	(void) strcpy(p, s);
	return ((int)p);
}

help(s)
	char *s;
{
	register struct tab *c;
	register int width, NCMDS;
#ifdef	CMUCS
	char tbuf[100];
#endif

	width = 0, NCMDS = 0;
	for (c = cmdtab; c->name != NULL; c++) {
#ifdef	CMUCS
		/* room for name, * and a space */
		int len = strlen(c->name) + 2;
#else
		int len = strlen(c->name);
#endif

#ifndef	CMUCS
		if (c->implemented == 0)
			len++;
#endif
		if (len > width)
			width = len;
		NCMDS++;
	}
	width = (width + 8) &~ 7;
	if (s == 0) {
		register int i, j, w;
		int columns, lines;

		lreply(214,
	  "The following commands are recognized (* =>'s unimplemented).");
		columns = 76 / width;
		if (columns == 0)
			columns = 1;
		lines = (NCMDS + columns - 1) / columns;
#ifdef	CMUCS
		c = cmdtab;
#endif
		for (i = 0; i < lines; i++) {
#ifdef	CMUCS
			*tbuf = '\0';
#else
			printf("   ");
#endif
			for (j = 0; j < columns; j++) {
				c = cmdtab + j * lines + i;
#ifdef	CMUCS
				strcat(tbuf, c->name);
				if (c->implemented)
				    strcat(tbuf, " ");
				else
				    strcat(tbuf, "*");
				w = strlen(c->name) + 1;
#else
				printf("%s%c", c->name,
					c->implemented ? ' ' : '*');
#endif
				if (c + lines >= &cmdtab[NCMDS])
					break;
#ifdef	CMUCS
				while (w++ < width)
					strcat(tbuf, " ");
#else
				w = strlen(c->name);
				while (w < width) {
					putchar(' ');
					w++;
				}
#endif
			}
#ifdef	CMUCS
			lreply(214," %s", tbuf);
#else
			printf("\r\n");
#endif
		}
		(void) fflush(stdout);
#ifdef	CMUCS
		reply(214, "Mail comments to Gripe@A.CS.CMU.EDU.");
#else
		reply(214, "Direct comments to ftp-bugs@%s.", hostname);
#endif
		return;
	}
	upper(s);
	c = lookup(s);
	if (c == (struct tab *)0) {
#ifdef	CMUCS
		reply(504, "Unknown help subject '%s'.", s);
#else
		reply(502, "Unknown command %s.", s);
#endif
		return;
	}
	if (c->implemented)
		reply(214, "Syntax: %s %s", c->name, c->help);
	else
		reply(214, "%-*s\t%s; unimplemented.", width, c->name, c->help);
}
