/*
 **********************************************************************
 * HISTORY
 * $Log:	ftpd.c,v $
 * Revision 1.1  92/04/09  01:31:26  vac
 * Initial revision
 * 
 * Revision 2.12  91/06/14  22:11:15  dlc
 * 	Fixed bug in user() when copying "anon" into the id variable when
 * 	id was "ftp", and therefor too small.
 * 	[91/06/14            dlc]
 * 
 * Revision 2.11  89/12/01  14:48:46  berman
 * 	Added check for null passwd around afs authentication message.
 * 	This prevents sending illegal 230 reply to any command that
 * 	causes a freelogin().
 * 	[89/12/01            berman]
 * 
 * Revision 2.10  89/10/17  13:11:50  berman
 * 	Added code under #if AFS to suspend the ALRM and PIPE signal
 * 	handlers for the duration of oklogin().  This allows oklogin to 
 * 	attempt to authenticate the user to AFS.
 * 	[89/10/17            berman]
 * 
 * Revision 2.9  89/09/12  18:29:10  mja
 * 	Use vsnprintf() instead of _doprnt().
 * 	[89/04/21            mja]
 * 
 * Revision 2.8  89/06/27  12:23:02  gm0w
 * 	Fixed small typo in previous edit.
 * 	[89/06/27            gm0w]
 * 
 * Revision 2.7  89/06/18  19:20:52  gm0w
 * 	Fixed to make hc happy.  Fixed to use varargs.
 * 	[89/06/18            gm0w]
 * 
 * Revision 2.6  89/06/18  15:10:50  gm0w
 * 	Changed local host comparison to make faster compare.
 * 	Removed setaid and account support code.
 * 	[89/06/18            gm0w]
 * 
 * Revision 2.5  89/01/30  14:18:06  dimitris
 * 	Modified chkacc to accept filenames starting with "/".
 * 
 * Revision 2.4  88/12/07  17:28:22  dimitris
 * 	Modified the login id assumed by guest from "ftpguest" to
 * 	"anon"
 * 	[88/12/07  17:22:34  dimitris]
 * 
 * Revision 2.3  88/11/14  10:44:04  dimitris
 * 	Changed the location of the local-hosts file to 
 * 		      /usr/cs/lib/hosts/local-hosts
 * 
 * Revision 2.2  88/11/08  14:47:39  dimitris
 * 	Changed code in order to compile under hc.
 * 	[88/11/08  14:40:42  dimitris]
 * 
 * Revision 2.1  88/10/28  15:37:26  dimitris
 * 	Add calls to allowed(filename) (in allowed.c) to let users 
 * 	open files or directories for read by ftpguest. Modified code to 
 * 	to allow ftpguest to login with username "anonymous"
 * 	or "ftp" and no password. (This assumes that the host
 * 	machine has an account for ftpguest.) Modified login() to
 * 	save remote host's address in rhostaddr wich oklocal() will
 * 	compare with addresses from /usr/cs/lib/local-host.
 * 	[88/10/28            dimitris]
 * 
 * 29-Aug-87  Rudy Nedved (ern) at Carnegie-Mellon University
 *	Add timeout code to receiving of data on data connection. Make
 *	/etc/unlogin a valid login shell for accounts like ftpguest.
 *
 * 11-Aug-87  Rudy Nedved (ern) at Carnegie-Mellon University
 *	fstat() on network file returns device of 0. This fools the 
 *	access code. Always set the device to -1 or unknown for 
 *	network stuff when calling the access code.
 *
 * 03-Nov-86  Jonathan J. Chew (jjc) at Carnegie-Mellon University
 *	Change dologin() to pass the net-to-host-long macro an address.
 *	
 * 24-May-86  Rudy Nedved (ern) at Carnegie-Mellon University
 *	Make idle timeout for real users be 30 minutes instead
 *	of 2 minutes. Tweak some things for 4.3BSD.
 *
 * 16-May-86  Rudy Nedved (ern) at Carnegie-Mellon University
 *	Explicit state the restrictions for guests and make the
 *	lack of a home directory non-fatal...chdir to root.
 *
 * 09-May-86  Rudy Nedved (ern) at Carnegie-Mellon University
 *	4.2BSD version with CMU modifications for free login,
 *	simple ftp guest login and logging code.
 *
 **********************************************************************
 */
/*
 * Copyright (c) 1985 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1985 Regents of the University of California.\n\
 All rights reserved.\n";
#endif not lint

#ifndef lint
static char sccsid[] = "@(#)ftpd.c	5.7 (Berkeley) 5/28/86";
#endif not lint

/*
 * FTP server.
 */
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/wait.h>

#include <netinet/in.h>

#include <arpa/ftp.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>

#include <stdio.h>
#include <signal.h>
#include <pwd.h>
#if	CMUCS
#include <grp.h>
#include <acc.h>
#include <sys/ttyloc.h>
#include <access.h>
#include <libc.h>
#endif	CMUCS
#include <setjmp.h>
#include <netdb.h>
#include <errno.h>
#include <strings.h>
#include <syslog.h>

#if	CMUCS
#define ANONUSER	"anon"
static char namebuf[128];
static char *userid = NULL;
static char *groupid = NULL;
#else	CMUCS
/*
 * File containing login names
 * NOT to be used on this machine.
 * Commonly used to disallow uucp.
 */
#define	FTPUSERS	"/etc/ftpusers"
#endif	CMUCS

extern  int allowed();
extern	int errno;
extern	char *sys_errlist[];
extern	char *crypt();
extern	char version[];
extern	char *home;		/* pointer to home directory for glob */
extern	FILE *popen(), *fopen(), *freopen();
extern	int  pclose(), fclose();
extern	char *getline();
extern	char cbuf[];

struct	sockaddr_in ctrl_addr;
struct	sockaddr_in data_source;
struct	sockaddr_in data_dest;
struct	sockaddr_in his_addr;

int	data;
jmp_buf	errcatch, urgcatch;
#if	CMUCS
jmp_buf	rcvcatch;
#endif	CMUCS
int	logged_in;
struct	passwd *pw;
#if	CMUCS
struct  group *gr;
#endif	CMUCS
int	debug;
#if	CMUCS
int	timeout = 900;
int	logging = 1;
#else	CMUCS
int	timeout = 900;    /* timeout after 15 minutes of inactivity */
int	logging;
#endif	CMUCS
int	guest;
int	wtmp;
int	type;
int	form;
int	stru;			/* avoid C keyword */
int	mode;
int	usedefault = 1;		/* for data transfers */
int	pdata;			/* for passive mode */
int	unique;
int	transflag;
char	tmpline[7];
#if	CMUCS
char	hostname[100];
char	remotehost[100];
char    rhostaddr[17];
#if	AFS
extern union wait afs_log_status; /* exit status or errno value */
extern int afs_log_errcode;
#define AFSERR_NOERR	0	/* no setup errors, see status */
#define AFSERR_NOPASS	1	/* no password, no authentication */
#define AFSERR_SIGUSED	2	/* signals in use, no authentication */
#define AFSERR_PIPE	3	/* pipe syscall, no authentication */
#define AFSERR_FORK	4	/* fork syscall, no authentication */
#define AFSERR_SETPAG	5	/* setpag syscall, no authentication */
#define AFSERR_TIMEOUT	6	/* timed out log, possible authentication */
#endif	/* AFS */
#else
char	hostname[32];
char	remotehost[32];
#endif	CMUCS

/*
 * Timeout intervals for retrying connections
 * to hosts that don't accept PORT cmds.  This
 * is a kludge, but given the problems with TCP...
 */
#define	SWAITMAX	90	/* wait at most 90 seconds */
#define	SWAITINT	5	/* interval between retries */

int	swaitmax = SWAITMAX;
int	swaitint = SWAITINT;

int	lostconn();
int	myoob();
FILE	*getdatasock(), *dataconn();

main(argc, argv)
	int argc;
	char *argv[];
{
	int addrlen, on = 1;
	long pgid;
	char *cp;

	addrlen = sizeof (his_addr);
	if (getpeername(0, &his_addr, &addrlen) < 0) {
		syslog(LOG_ERR, "getpeername (%s): %m",argv[0]);
		exit(1);
	}
	addrlen = sizeof (ctrl_addr);
	if (getsockname(0, (char *) &ctrl_addr, &addrlen) < 0) {
		syslog(LOG_ERR, "getsockname (%s): %m",argv[0]);
		exit(1);
	}
	data_source.sin_port = htons(ntohs(ctrl_addr.sin_port) - 1);
	debug = 0;
	openlog("ftpd", LOG_PID, LOG_DAEMON);
	argc--, argv++;
	while (argc > 0 && *argv[0] == '-') {
		for (cp = &argv[0][1]; *cp; cp++) switch (*cp) {

		case 'v':
			debug = 1;
			break;

		case 'd':
			debug = 1;
			break;

		case 'l':
			logging = 1;
			break;

		case 't':
			timeout = atoi(++cp);
#if	CMUCS
			/* break has same effect */
#else	CMUCS
			goto nextopt;
#endif	CMUCS
			break;

		default:
			fprintf(stderr, "ftpd: Unknown flag -%c ignored.",
			     *cp);
			break;
		}
#if	CMUCS
	/* not needed */
#else	CMUCS
nextopt:
#endif	CMUCS
		argc--, argv++;
	}
	(void) signal(SIGPIPE, lostconn);
	(void) signal(SIGCHLD, SIG_IGN);
	if ((int) signal(SIGURG, myoob) < 0) {
		syslog(LOG_ERR, "signal: %m");
	}
	/* handle urgent data inline */
#ifdef SO_OOBINLINE
	if (setsockopt(0, SOL_SOCKET, SO_OOBINLINE, (char *)&on, sizeof(on)) < 0) {
		syslog(LOG_ERR, "setsockopt: %m");
	}
#endif SO_OOBINLINE
	pgid = getpid();
	if (ioctl(fileno(stdin), SIOCSPGRP, (char *) &pgid) < 0) {
		syslog(LOG_ERR, "ioctl: %m");
	}
#ifndef VACFASTFTPD
	dolog(&his_addr);
#endif
	/* do telnet option negotiation here */
	/*
	 * Set up default state
	 */
	logged_in = 0;
	data = -1;
	type = TYPE_A;
	form = FORM_N;
	stru = STRU_F;
	mode = MODE_S;
	tmpline[0] = '\0';
	(void) gethostname(hostname, sizeof (hostname));
	reply(220, "%s FTP server (%s) ready.",
		hostname, version);
	for (;;) {
		(void) setjmp(errcatch);
		(void) yyparse();
	}
}

lostconn()
{

#if	CMUCS
	if (logging)
#else	CMUCS
	if (debug)
#endif	CMUCS
		syslog(LOG_DEBUG, "lost connection");
#ifndef VACFASTFTPD
	dologout(-1);
#endif
}

pass(passwd)
	char *passwd;
{
#if	CMUCS
	/* CMU CS uses oklogin() */
#else	CMUCS
	char *xpasswd, *savestr();
	static struct passwd save;
#endif	CMUCS

#if	CMUCS
	if (logged_in) {
		reply(503, "You are already logged in.");
		return;
	}
	if (userid == 0) {
#else	CMUCS
	if (logged_in || pw == NULL) {
#endif	CMUCS
		reply(503, "Login with USER first.");
		return;
	}
#if	CMUCS
#ifndef VACFASTFTPD
	if (dologin(passwd) == 0)
		return;
#endif
	if (!guest)
		reply(230, "User %s logged in.", pw->pw_name);
	else {
		reply(230, "Filenames can not have '/..' in them.");
	}
#else	CMUCS
	if (!guest) {		/* "ftp" is only account allowed no password */
		xpasswd = crypt(passwd, pw->pw_passwd);
		/* The strcmp does not catch null passwords! */
		if (*pw->pw_passwd == '\0' || strcmp(xpasswd, pw->pw_passwd)) {
			reply(530, "Login incorrect.");
			pw = NULL;
			return;
		}
	}
	setegid(pw->pw_gid);
	initgroups(pw->pw_name, pw->pw_gid);
	if (chdir(pw->pw_dir)) {
		reply(530, "User %s: can't change directory to %s.",
			pw->pw_name, pw->pw_dir);
		goto bad;
	}

	/* grab wtmp before chroot */
	wtmp = open("/usr/adm/wtmp", O_WRONLY|O_APPEND);
	if (guest && chroot(pw->pw_dir) < 0) {
		reply(550, "Can't set guest privileges.");
		if (wtmp >= 0) {
			(void) close(wtmp);
			wtmp = -1;
		}
		goto bad;
	}
	if (!guest)
		reply(230, "User %s logged in.", pw->pw_name);
	else
		reply(230, "Guest logged in.");
	logged_in = 1;
#ifndef VACFASTFTPD
	dologin(pw);
#endif
	seteuid(pw->pw_uid);
	/*
	 * Save everything so globbing doesn't
	 * clobber the fields.
	 */
	save = *pw;
	save.pw_name = savestr(pw->pw_name);
	save.pw_passwd = savestr(pw->pw_passwd);
	save.pw_comment = savestr(pw->pw_comment);
	save.pw_gecos = savestr(pw->pw_gecos);
	save.pw_dir = savestr(pw->pw_dir);
	save.pw_shell = savestr(pw->pw_shell);
	pw = &save;
	home = pw->pw_dir;		/* home dir for globbing */
	return;
bad:
	seteuid(0);
	pw = NULL;
#endif	CMUCS
}


char *
savestr(s)
	char *s;
{
	char *malloc();

	char *new = malloc((unsigned) strlen(s) + 1);
	
	if (new != NULL)
		(void) strcpy(new, s);
	return (new);
}

retrieve(cmd, name)
	char *cmd, *name;
{
	FILE *fin, *dout;
	struct stat st;
	int (*closefunc)(), tmp;

#if	CMUCS
	if (guest && (!allowed(name))) {
	   reply(530, "The file %s is not accessible to guest users",name);
	    return;
	}
#endif	CMUCS
	if (cmd == 0) {
#ifdef notdef
		/* no remote command execution -- it's a security hole */
		if (*name == '|')
			fin = popen(name + 1, "r"), closefunc = pclose;
		else
#endif
			fin = fopen(name, "r"), closefunc = fclose;
	} else {
		char line[BUFSIZ];

		(void) sprintf(line, cmd, name), name = line;
		fin = popen(line, "r"), closefunc = pclose;
	}
	if (fin == NULL) {
		if (errno != 0)
			reply(550, "%s: %s.", name, sys_errlist[errno]);
		return;
	}
#if	CMUCS
	st.st_size = (off_t)-1;
#else	CMUCS
	st.st_size = 0;
#endif	CMUCS
	if (cmd == 0 &&
	    (stat(name, &st) < 0 || (st.st_mode&S_IFMT) != S_IFREG)) {
		reply(550, "%s: not a plain file.", name);
		goto done;
	}
	dout = dataconn(name, st.st_size, "w");
	if (dout == NULL)
		goto done;
	if ((tmp = send_data(fin, dout)) > 0 || ferror(dout) > 0) {
		reply(550, "%s: %s.", name, sys_errlist[errno]);
	}
	else if (tmp == 0) {
		reply(226, "Transfer complete.");
	}
	(void) fclose(dout);
	data = -1;
	pdata = -1;
done:
	(*closefunc)(fin);
}

store(name, mode)
	char *name, *mode;
{
	FILE *fout, *din;
	int (*closefunc)(), dochown = 0, tmp;
	char *gunique(), *local;

#if	CMUCS
	if (guest && (!allowed(name))) {
	   reply(530, "The file %s is no accessible to for guest users",name);
	    return;
	}
#endif	CMUCS
#ifdef notdef
	/* no remote command execution -- it's a security hole */
	if (name[0] == '|')
		fout = popen(&name[1], "w"), closefunc = pclose;
	else
#endif
	{
		struct stat st;

		local = name;
		if (stat(name, &st) < 0) {
			dochown++;
		}
		else if (unique) {
			if ((local = gunique(name)) == NULL) {
				return;
			}
			dochown++;
		}
		fout = fopen(local, mode), closefunc = fclose;
	}
	if (fout == NULL) {
		reply(553, "%s: %s.", local, sys_errlist[errno]);
		return;
	}
	din = dataconn(local, (off_t)-1, "r");
	if (din == NULL)
		goto done;
	if ((tmp = receive_data(din, fout)) > 0 || ferror(fout) > 0) {
		reply(552, "%s: %s.", local, sys_errlist[errno]);
	}
	else if (tmp == 0 && !unique) {
		reply(226, "Transfer complete.");
	}
	else if (tmp == 0 && unique) {
		reply(226, "Transfer complete (unique file name:%s).", local);
	}
	(void) fclose(din);
	data = -1;
	pdata = -1;
done:
	if (dochown)
		(void) chown(local, pw->pw_uid, -1);
	(*closefunc)(fout);
}

FILE *
getdatasock(mode)
	char *mode;
{
	int s, on = 1;

	if (data >= 0)
		return (fdopen(data, mode));
	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0)
		return (NULL);
	seteuid(0);
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof (on)) < 0)
		goto bad;
	/* anchor socket to avoid multi-homing problems */
	data_source.sin_family = AF_INET;
	data_source.sin_addr = ctrl_addr.sin_addr;
	if (bind(s, &data_source, sizeof (data_source)) < 0)
		goto bad;
	seteuid(pw->pw_uid);
	return (fdopen(s, mode));
bad:
	seteuid(pw->pw_uid);
	(void) close(s);
	return (NULL);
}

FILE *
dataconn(name, size, mode)
	char *name;
	off_t size;
	char *mode;
{
	char sizebuf[32];
	FILE *file;
	int retry = 0;

	if (size >= 0)
		(void) sprintf (sizebuf, " (%ld bytes)", size);
	else
		(void) strcpy(sizebuf, "");
	if (pdata > 0) {
		struct sockaddr_in from;
		int s, fromlen = sizeof(from);

		s = accept(pdata, &from, &fromlen);
		if (s < 0) {
			reply(425, "Can't open data connection.");
			(void) close(pdata);
			pdata = -1;
			return(NULL);
		}
		(void) close(pdata);
		pdata = s;
		reply(150, "Openning data connection for %s (%s,%d)%s.",
		     name, inet_ntoa(from.sin_addr),
		     ntohs(from.sin_port), sizebuf);
		return(fdopen(pdata, mode));
	}
	if (data >= 0) {
		reply(125, "Using existing data connection for %s%s.",
		    name, sizebuf);
		usedefault = 1;
		return (fdopen(data, mode));
	}
	if (usedefault)
		data_dest = his_addr;
	usedefault = 1;
	file = getdatasock(mode);
	if (file == NULL) {
		reply(425, "Can't create data socket (%s,%d): %s.",
		    inet_ntoa(data_source.sin_addr),
		    ntohs(data_source.sin_port),
		    sys_errlist[errno]);
		return (NULL);
	}
	data = fileno(file);
	while (connect(data, &data_dest, sizeof (data_dest)) < 0) {
		if (errno == EADDRINUSE && retry < swaitmax) {
			sleep((unsigned) swaitint);
			retry += swaitint;
			continue;
		}
		reply(425, "Can't build data connection: %s.",
		    sys_errlist[errno]);
		(void) fclose(file);
		data = -1;
		return (NULL);
	}
	reply(150, "Opening data connection for %s (%s,%d)%s.",
	    name, inet_ntoa(data_dest.sin_addr),
	    ntohs(data_dest.sin_port), sizebuf);
	return (file);
}

/*
 * Tranfer the contents of "instr" to
 * "outstr" peer using the appropriate
 * encapulation of the date subject
 * to Mode, Structure, and Type.
 *
 * NB: Form isn't handled.
 */
send_data(instr, outstr)
	FILE *instr, *outstr;
{
	register int c;
	int netfd, filefd, cnt;
	char buf[BUFSIZ];

	transflag++;
	if (setjmp(urgcatch)) {
		transflag = 0;
		return(-1);
	}
	switch (type) {

	case TYPE_A:
		while ((c = getc(instr)) != EOF) {
			if (c == '\n') {
				if (ferror (outstr)) {
					transflag = 0;
					return (1);
				}
				(void) putc('\r', outstr);
			}
			(void) putc(c, outstr);
		/*	if (c == '\r')			*/
		/*		putc ('\0', outstr);	*/
		}
		transflag = 0;
		if (ferror (instr) || ferror (outstr)) {
			return (1);
		}
		return (0);
		
	case TYPE_I:
	case TYPE_L:
		netfd = fileno(outstr);
		filefd = fileno(instr);

		while ((cnt = read(filefd, buf, sizeof (buf))) > 0) {
			if (write(netfd, buf, cnt) < 0) {
				transflag = 0;
				return (1);
			}
		}
		transflag = 0;
		return (cnt < 0);
	}
	reply(550, "Unimplemented TYPE %d in send_data", type);
	transflag = 0;
	return (-1);
}

#if	CMUCS
rcvtimout()
{
	/* only process if transfer occurring */
	if (!transflag) {
		return;
	}
	alarm(0);
	reply(426,"Receive data timeout. Transfer aborted. Data connection closed.");
	longjmp(rcvcatch, 1);
}
#define	RCVTIMOUT	(5*60)
#endif	CMUCS

/*
 * Transfer data from peer to
 * "outstr" using the appropriate
 * encapulation of the data subject
 * to Mode, Structure, and Type.
 *
 * N.B.: Form isn't handled.
 */
receive_data(instr, outstr)
	FILE *instr, *outstr;
{
	register int c;
	int cnt;
	char buf[BUFSIZ];


	transflag++;
	if (setjmp(urgcatch)) {
		transflag = 0;
		return(-1);
	}
#if	CMUCS
	(void) signal(SIGALRM, rcvtimout);
	if (setjmp(rcvcatch)) {
		transflag = 0;
		return(-1);
	}
#endif	CMUCS
	switch (type) {

	case TYPE_I:
	case TYPE_L:
#if	CMUCS
		for(;;) {
			alarm(RCVTIMOUT);
			cnt = read(fileno(instr), buf, sizeof buf);
			alarm(0);
			if (cnt <= 0) break;
#else	CMUCS
		while ((cnt = read(fileno(instr), buf, sizeof buf)) > 0) {
#endif	CMUCS
			if (write(fileno(outstr), buf, cnt) < 0) {
				transflag = 0;
				return (1);
			}
		}
		transflag = 0;
		return (cnt < 0);

	case TYPE_E:
		reply(553, "TYPE E not implemented.");
		transflag = 0;
		return (-1);

	case TYPE_A:
#if	CMUCS
		for(;;) {
			if (instr->_cnt <= 0) {
			    alarm(RCVTIMOUT);
			    c = getc(instr);
			    alarm(0);
			}
			else
			    c = getc(instr);
			if (c == EOF) break;
#else	CMUCS
		while ((c = getc(instr)) != EOF) {
#endif	CMUCS
			while (c == '\r') {
				if (ferror (outstr)) {
					transflag = 0;
					return (1);
				}
				if ((c = getc(instr)) != '\n')
					(void) putc ('\r', outstr);
			/*	if (c == '\0')			*/
			/*		continue;		*/
			}
			(void) putc (c, outstr);
		}
		transflag = 0;
		if (ferror (instr) || ferror (outstr))
			return (1);
		return (0);
	}
	transflag = 0;
	fatal("Unknown type in receive_data.");
	/*NOTREACHED*/
}

fatal(s)
	char *s;
{
#if	CMUCS
	reply(421, "Closing connection due to server error: %s", s);
#else	CMUCS
	reply(451, "Error in server: %s", s);
	reply(221, "Closing connection due to server error.");
#endif	CMUCS
#ifndef VACFASTFTPD
	dologout(0);
#endif
}

#if	CMUCS
/* new versions of these appear at the end of the file */
#else	CMUCS
/*VARARGS2*/
reply(n, s, args)
	int n;
	char *s;
{

	printf("%d ", n);
	_doprnt(s, &args, stdout);
	printf("\r\n");
	(void) fflush(stdout);
	if (debug) {
		syslog(LOG_DEBUG, "<--- %d ", n);
		syslog(LOG_DEBUG, s, &args);
	}
}

/*VARARGS2*/
lreply(n, s, args)
	int n;
	char *s;
{
	printf("%d-", n);
	_doprnt(s, &args, stdout);
	printf("\r\n");
	(void) fflush(stdout);
	if (debug) {
		syslog(LOG_DEBUG, "<--- %d- ", n);
		syslog(LOG_DEBUG, s, &args);
	}
}
#endif	CMUCS

ack(s)
	char *s;
{
	reply(250, "%s command successful.", s);
}

nack(s)
	char *s;
{
	reply(502, "%s command not implemented.", s);
}

yyerror(s)
	char *s;
{
	char *cp;

	cp = index(cbuf,'\n');
	*cp = '\0';
	reply(500, "'%s': command not understood.",cbuf);
}

delete(name)
	char *name;
{
	struct stat st;

#if	CMUCS
	if (guest) {
	    reply(530, "Access not allowed for guest users");
	    return;
	}
#endif	CMUCS
	if (stat(name, &st) < 0) {
		reply(550, "%s: %s.", name, sys_errlist[errno]);
		return;
	}
	if ((st.st_mode&S_IFMT) == S_IFDIR) {
		if (rmdir(name) < 0) {
			reply(550, "%s: %s.", name, sys_errlist[errno]);
			return;
		}
		goto done;
	}
	if (unlink(name) < 0) {
		reply(550, "%s: %s.", name, sys_errlist[errno]);
		return;
	}
done:
#if	CMUCS
	reply(250, "File %s deleted.", name);
#else	CMUCS
	ack("DELE");
#endif	CMUCS
}

cwd(path)
	char *path;
{

#if	CMUCS
	if (guest && (!allowed(path))) {
	    reply(530, "Access not allowed for guest users for path %s",path);
	    return;
	}
#endif	CMUCS
	if (chdir(path) < 0) {
		reply(550, "%s: %s.", path, sys_errlist[errno]);
		return;
	}
#if	CMUCS
	abspath(0, 0);
	reply(250, "Directory path set to %s.", path);
#else	CMUCS
	ack("CWD");
#endif	CMUCS
}

#if	CMUCS
char *quoteit(nam)
	register char *nam;
{
	static char mybuf[BUFSIZ];
	register char *p, c;

	p = mybuf;
	while ((c = *p++ = *nam++) != '\0')
		if (c == '"')
			*p++ = '"';
	return(mybuf);
}
#endif	CMUCS

makedir(name)
	char *name;
{
	struct stat st;
	int dochown = stat(name, &st) < 0;
#if	CMUCS
	char tbuf[BUFSIZ];
#endif	CMUCS
	
#if	CMUCS
	if (guest) {
	    reply(530, "Access not allowed for guest users");
	    return;
	}
#endif	CMUCS
	if (mkdir(name, 0777) < 0) {
		reply(550, "%s: %s.", name, sys_errlist[errno]);
		return;
	}
	if (dochown)
		(void) chown(name, pw->pw_uid, -1);
#if	CMUCS
	abspath(name, tbuf);
	reply(257, "Directory %s created." ,quoteit(tbuf));
#else	CMUCS
	reply(257, "MKD command successful.");
#endif	CMUCS
}

removedir(name)
	char *name;
{

#if	CMUCS
	if (guest) {
	    reply(530, "Access not allowed for guest users");
	    return;
	}
#endif	CMUCS
	if (rmdir(name) < 0) {
		reply(550, "%s: %s.", name, sys_errlist[errno]);
		return;
	}
#if	CMUCS
	reply(250, "Directory %s deleted.", name);
#else	CMUCS
	ack("RMD");
#endif	CMUCS
}

pwd()
{
	char path[MAXPATHLEN + 1];

	if (getwd(path) == NULL) {
		reply(550, "%s.", path);
		return;
	}
#if	CMUCS
	reply(257, "\"%s\" is current directory.", quoteit(path));
#else	CMUCS
	reply(257, "\"%s\" is current directory.", path);
#endif	CMUCS
}

char *
renamefrom(name)
	char *name;
{
	struct stat st;

#if	CMUCS
	if (guest) {
	    reply(530, "Access not allowed for guest users");
	    return ((char *)0);
	}
#endif	CMUCS
	if (stat(name, &st) < 0) {
		reply(550, "%s: %s.", name, sys_errlist[errno]);
		return ((char *)0);
	}
	reply(350, "File exists, ready for destination name");
	return (name);
}

renamecmd(from, to)
	char *from, *to;
{

#if	CMUCS
	if (guest) {
	    reply(530, "Access not allowed for guest users");
	    return;
	}
#endif	CMUCS
	if (rename(from, to) < 0) {
		reply(550, "rename: %s.", sys_errlist[errno]);
		return;
	}
#if	CMUCS
	reply(250, "Renamed %s to %s.", from, to);
#else	CMUCS
	ack("RNTO");
#endif	CMUCS
}


#ifndef VACFASTFTPD

dolog(sin)
	struct sockaddr_in *sin;
{

	struct hostent *hp = gethostbyaddr(&sin->sin_addr,
		sizeof (struct in_addr), AF_INET);
	time_t t;
	extern char *ctime();

#if CMUCS
        (void) strncpy(rhostaddr,inet_ntoa(sin->sin_addr),sizeof(rhostaddr));
#endif CMUCS
	if (hp) {
		(void) strncpy(remotehost, hp->h_name, sizeof (remotehost));
		endhostent();
	} else
#if	CMUCS
		sprintf(remotehost, "[%s]",inet_ntoa(sin->sin_addr));
#else
		(void) strncpy(remotehost, inet_ntoa(sin->sin_addr),
		    sizeof (remotehost));
#endif	CMUCS
	if (!logging)
		return;
	t = time((time_t *) 0);
	syslog(LOG_INFO,"FTPD: connection from %s at %s", remotehost, ctime(&t));
}

#if	CMUCS
dologout(status)
{
	if (logging)
		syslog(LOG_INFO, "server exiting with status %#o", status);

	/* beware of flushing buffers after a SIGPIPE */
	_exit(status);
}
#else	CMUCS
#include <utmp.h>

#define	SCPYN(a, b)	(void) strncpy(a, b, sizeof (a))
struct	utmp utmp;

/*
 * Record login in wtmp file.
 */
dologin(pw)
	struct passwd *pw;
{
	char line[32];

	if (wtmp >= 0) {
		/* hack, but must be unique and no tty line */
		(void) sprintf(line, "ftp%d", getpid());
		SCPYN(utmp.ut_line, line);
		SCPYN(utmp.ut_name, pw->pw_name);
		SCPYN(utmp.ut_host, remotehost);
		utmp.ut_time = (long) time((time_t *) 0);
		(void) write(wtmp, (char *)&utmp, sizeof (utmp));
		if (!guest) {		/* anon must hang on */
			(void) close(wtmp);
			wtmp = -1;
		}
	}
}

/*
 * Record logout in wtmp file
 * and exit with supplied status.
 */
dologout(status)
	int status;
{

	if (logged_in) {
		(void) seteuid(0);
		if (wtmp < 0)
			wtmp = open("/usr/adm/wtmp", O_WRONLY|O_APPEND);
		if (wtmp >= 0) {
			SCPYN(utmp.ut_name, "");
			SCPYN(utmp.ut_host, "");
			utmp.ut_time = (long) time((time_t *) 0);
			(void) write(wtmp, (char *)&utmp, sizeof (utmp));
			(void) close(wtmp);
		}
	}
	/* beware of flushing buffers after a SIGPIPE */
	_exit(status);
}
#endif

#endif         /* VACFASTFTPD */

/*
 * Special version of popen which avoids
 * call to shell.  This insures noone may 
 * create a pipe to a hidden program as a side
 * effect of a list or dir command.
 */
#define	tst(a,b)	(*mode == 'r'? (b) : (a))
#define	RDR	0
#define	WTR	1
static	int popen_pid[5];

static char *
nextarg(cpp)
	char *cpp;
{
	register char *cp = cpp;

	if (cp == 0)
		return (cp);
	while (*cp && *cp != ' ' && *cp != '\t')
		cp++;
	if (*cp == ' ' || *cp == '\t') {
		*cp++ = '\0';
		while (*cp == ' ' || *cp == '\t')
			cp++;
	}
	if (cp == cpp)
		return ((char *)0);
	return (cp);
}

FILE *
popen(cmd, mode)
	char *cmd, *mode;
{
	int p[2], ac, gac;
	register myside, hisside, pid;
	char *av[20], *gav[512];
	register char *cp;

	if (pipe(p) < 0)
		return (NULL);
	cp = cmd, ac = 0;
	/* break up string into pieces */
	do {
		av[ac++] = cp;
		cp = nextarg(cp);
	} while (cp && *cp && ac < 20);
	av[ac] = (char *)0;
	gav[0] = av[0];
	/* glob each piece */
	for (gac = ac = 1; av[ac] != NULL; ac++) {
		char **pop;
		extern char **glob(), **copyblk();

		pop = glob(av[ac]);
		if (pop == (char **)NULL) {	/* globbing failed */
			char *vv[2];

			vv[0] = av[ac];
			vv[1] = 0;
			pop = copyblk(vv);
		}
		av[ac] = (char *)pop;		/* save to free later */
		while (*pop && gac < 512)
			gav[gac++] = *pop++;
	}
	gav[gac] = (char *)0;
	myside = tst(p[WTR], p[RDR]);
	hisside = tst(p[RDR], p[WTR]);
#if	CMUCS
	pid = -1;
	if (guest)
		for (gac = 1; gav[gac] != 0; gac++)
			if (chkacc(gav[gac])) {
				errno = EACCES;
				goto skipfork;
			}
#endif	CMUCS
	if ((pid = fork()) == 0) {
		/* myside and hisside reverse roles in child */
		(void) close(myside);
		(void) dup2(hisside, tst(0, 1));
		(void) close(hisside);
#if	CMUCS
		execvp(gav[0], gav);
#else	CMUCS
		execv(gav[0], gav);
#endif	CMUCS
		_exit(1);
	}
#if	CMUCS
skipfork:
#endif	CMUCS
	for (ac = 1; av[ac] != NULL; ac++)
		blkfree((char **)av[ac]);
	if (pid == -1)
		return (NULL);
	popen_pid[myside] = pid;
	(void) close(hisside);
	return (fdopen(myside, mode));
}

pclose(ptr)
	FILE *ptr;
{
	register f, r, (*hstat)(), (*istat)(), (*qstat)();
	int status;

	f = fileno(ptr);
	(void) fclose(ptr);
	istat = signal(SIGINT, SIG_IGN);
	qstat = signal(SIGQUIT, SIG_IGN);
	hstat = signal(SIGHUP, SIG_IGN);
	while ((r = wait(&status)) != popen_pid[f] && r != -1)
		;
	if (r == -1)
		status = -1;
	(void) signal(SIGINT, istat);
	(void) signal(SIGQUIT, qstat);
	(void) signal(SIGHUP, hstat);
	return (status);
}

#if	CMUCS
/* CMU CS uses oklogin() */
#else	CMUCS
/*
 * Check user requesting login priviledges.
 * Disallow anyone who does not have a standard
 * shell returned by getusershell() (/etc/shells).
 * Disallow anyone mentioned in the file FTPUSERS
 * to allow people such as uucp to be avoided.
 */
checkuser(name)
	register char *name;
{
	register char *cp;
	char line[BUFSIZ], *index(), *getusershell();
	FILE *fd;
	struct passwd *pw;
	int found = 0;

	pw = getpwnam(name);
	if (pw == NULL)
		return (0);
	while ((cp = getusershell()) != NULL)
		if (strcmp(cp, pw->pw_shell) == 0)
			break;
	endpwent();
	endusershell();
	if (cp == NULL)
		return (0);
	fd = fopen(FTPUSERS, "r");
	if (fd == NULL)
		return (1);
	while (fgets(line, sizeof (line), fd) != NULL) {
		cp = index(line, '\n');
		if (cp)
			*cp = '\0';
		if (strcmp(line, name) == 0) {
			found++;
			break;
		}
	}
	(void) fclose(fd);
	return (!found);
}
#endif	CMUCS

myoob()
{
	char *cp;

	/* only process if transfer occurring */
	if (!transflag) {
		return;
	}
	cp = tmpline;
	if (getline(cp, 7, stdin) == NULL) {
		reply(221, "You could at least say goodby.");
#ifndef VACFASTFTPD
		dologout(0);
#endif
	}
	upper(cp);
	if (strcmp(cp, "ABOR\r\n"))
		return;
	tmpline[0] = '\0';
	reply(426,"Transfer aborted. Data connection closed.");
	reply(226,"Abort successful");
	longjmp(urgcatch, 1);
}

/*
 * Note: The 530 reply codes could be 4xx codes, except nothing is
 * given in the state tables except 421 which implies an exit.  (RFC959)
 */
passive()
{
	int len;
	struct sockaddr_in tmp;
	register char *p, *a;

	pdata = socket(AF_INET, SOCK_STREAM, 0);
	if (pdata < 0) {
		reply(530, "Can't open passive connection");
		return;
	}
	tmp = ctrl_addr;
	tmp.sin_port = 0;
	seteuid(0);
	if (bind(pdata, (struct sockaddr *) &tmp, sizeof(tmp)) < 0) {
		seteuid(pw->pw_uid);
		(void) close(pdata);
		pdata = -1;
		reply(530, "Can't open passive connection");
		return;
	}
	seteuid(pw->pw_uid);
	len = sizeof(tmp);
	if (getsockname(pdata, (char *) &tmp, &len) < 0) {
		(void) close(pdata);
		pdata = -1;
		reply(530, "Can't open passive connection");
		return;
	}
	if (listen(pdata, 1) < 0) {
		(void) close(pdata);
		pdata = -1;
		reply(530, "Can't open passive connection");
		return;
	}
	a = (char *) &tmp.sin_addr;
	p = (char *) &tmp.sin_port;

#define UC(b) (((int) b) & 0xff)

	reply(227, "Entering Passive Mode (%d,%d,%d,%d,%d,%d)", UC(a[0]),
		UC(a[1]), UC(a[2]), UC(a[3]), UC(p[0]), UC(p[1]));
}

char *
gunique(local)
	char *local;
{
	static char new[MAXPATHLEN];
	char *cp = rindex(local, '/');
	int d, count=0;
	char ext = '1';

	if (cp) {
		*cp = '\0';
	}
	d = access(cp ? local : ".", 2);
	if (cp) {
		*cp = '/';
	}
	if (d < 0) {
		syslog(LOG_ERR, "%s: %m", local);
		return((char *) 0);
	}
	(void) strcpy(new, local);
	cp = new + strlen(new);
	*cp++ = '.';
	while (!d) {
		if (++count == 100) {
			reply(452, "Unique file name not cannot be created.");
			return((char *) 0);
		}
		*cp++ = ext;
		*cp = '\0';
		if (ext == '9') {
			ext = '0';
		}
		else {
			ext++;
		}
		if ((d = access(new, 0)) < 0) {
			break;
		}
		if (ext != '0') {
			cp--;
		}
		else if (*(cp - 2) == '.') {
			*(cp - 1) = '1';
		}
		else {
			*(cp - 2) = *(cp - 2) + 1;
			cp--;
		}
	}
	return(new);
}

#if	CMUCS
user(id)
	char *id;
{
	if (logged_in) {
		reply(503, "You are already logged in.");
		return;
	}
        if (guest)
		id = "anon";

        strcpy(namebuf, id);
	userid = namebuf;

	/*
	 *  Pick up any group specified after a comma in the
	 *  login name.  Treat a null group as if none were
	 *  specified.
	 */
	if ((groupid=index(userid, ',')) != 0) {
		*groupid++ = 0;
/* begin temporary account phase-out support */
		if ((userid = index(groupid, ',')) != 0)
			*userid = 0;
		userid = namebuf;
/* end temporary account phase-out support */
		if (*groupid == 0)
			groupid = 0;
	}
        if (!guest)
        	reply(331, "Password required for %s.", id);
}

/*
 * login using the password string (NULL for no password)
 */
dologin(passwd)
char *passwd;
{
	char *savestr(), *okpassword(), *getusershell();
	static struct passwd save;
	int code, val;
	struct stat st;
	struct ttyloc tlc;
	int *grlist;
	char *p;

	/* make sure this is set right */
	logged_in = 0;

	/* see if guest or not */

	/* get some information for access logging */
	st.st_rdev = (dev_t)-1;	/* fstat returns 0 */
	tlc.tlc_hostid = ntohl(his_addr.sin_addr.s_addr);
	tlc.tlc_ttyid = TLC_UNKTTY;
	gr = NULL;

	if ((pw = getpwnam(userid)) == NULL)
		code = ACCESS_CODE_NOUSER;
	else {
		code = okaccess(userid, ACCESS_TYPE_FTP, getuid(), st.st_rdev, tlc);
		code = (code == 1) ? ACCESS_CODE_OK : ACCESS_CODE_DENIED;
                if (guest)
                   passwd=NULL; /* This tells oklogin() not to do 
                                   password checking */
		{
			char *accountid = NULL;
			struct account *ac = NULL;

#if	AFS
			/*
			 * Temporarily remove the signal handlers for 
			 * these signals since oklogin() will need them
			 * to be NULL to give AFS authentication.
			 * Also save the alarm clock so that it can be
			 * restored after oklogin().
			 */
			struct sigvec	dum_vec, oalrm_vec, opipe_vec;
			unsigned int	alarm_time;

#define 		NULLVEC (struct sigvec *)NULL

			dum_vec.sv_handler = SIG_IGN;
			alarm_time = alarm((unsigned) 0);
		        (void) sigvec(SIGALRM, &dum_vec, &oalrm_vec);
		        (void) sigvec(SIGPIPE, &dum_vec, &opipe_vec);
#endif	/* AFS */
	                val = oklogin(userid, groupid, &accountid, passwd,
				      &pw, &gr, &ac, &grlist);
#if	AFS
		        (void) sigvec(SIGALRM, &oalrm_vec, NULLVEC);
		        (void) sigvec(SIGPIPE, &opipe_vec, NULLVEC);
			(void) alarm(alarm_time);
#endif	/* AFS */
		}
		if (val != ACCESS_CODE_OK)
			code = val;
	}
        if (guest) code = ACCESS_CODE_OK; 
	switch (code) {
	case ACCESS_CODE_OK:		/* everything checks out */
		p = okpassword(passwd, pw->pw_name, pw->pw_gecos);
		if ((!guest) && (p != NULL)) {
			lreply(530, "%s", p);
			code = ACCESS_CODE_INSECUREPWD;
		} else {
			if (pw->pw_passwd[0] != '\0') {
				if (pw->pw_shell[0] == '\0')
					break;
				while ((p = getusershell()) != NULL)
					if (strcmp(p, pw->pw_shell) == 0)
						break;
				endusershell();
				if (p != 0)
					break;
				/* the following is special for ftp */
				if (strcmp(pw->pw_shell,"/etc/unlogin") == 0)
				    break;
			}
			code = ACCESS_CODE_DENIED;
		}
	default:
	case ACCESS_CODE_DENIED:
	case ACCESS_CODE_NOUSER:
	case ACCESS_CODE_BADPASSWORD:
		reply(530, "Login incorrect.");
		break;
	case ACCESS_CODE_ACCEXPIRED:
		reply(530, "Sorry, your account(s) have expired.");
		break;
	case ACCESS_CODE_GRPEXPIRED:
		reply(530, "Sorry, your %s group account has expired",
			foldup(gr->gr_name, gr->gr_name));
		break;
	case ACCESS_CODE_ACCNOTVALID:
		reply(530, "Sorry, %s is not a valid account group",
			foldup(gr->gr_name, gr->gr_name));
		break;
	case ACCESS_CODE_MANYDEFACC:
		lreply(530, "Your accounting file entries are inconsistant.");
		reply(530, "Please report this problem to Gripe.");
		break;
	case ACCESS_CODE_NOACCFORGRP:
		reply(530, "Sorry, your %s group does not have an account.",
			foldup(gr->gr_name, gr->gr_name));
		break;
	case ACCESS_CODE_NOGRPFORACC:
		reply(530, "Sorry, your account(s) do not have group(s).");
		break;
	case ACCESS_CODE_NOGRPDEFACC:
		reply(530, "Sorry, your default account does not have a group.");
		break;
	case ACCESS_CODE_NOTGRPMEMB:
		reply(530, "Sorry, you are not a member of the %s group",
			foldup(groupid, groupid));
		break;
	case ACCESS_CODE_NOTDEFMEMB:
		lreply(530, "Sorry, you are not a member of your default login");
		reply(530, "group.  You must specify an explicit group to login");
		break;
	}

	/* log access */
	logaccess(userid, ACCESS_TYPE_FTP, code, getuid(), st.st_rdev, tlc);

	if (code != ACCESS_CODE_OK) {
		userid = NULL;
		pw = NULL;
		gr = NULL;
		return 0;
	}
#if 	AFS
	if (code == ACCESS_CODE_OK && passwd != NULL) {
		switch (afs_log_errcode) {
		case AFSERR_NOERR: /* no setup errors, see status */
			if (afs_log_status.w_stopval == WSTOPPED) {
				lreply(230, "[ %s: process stopped ]",
					"AFS authentication failed");
				break;
			}
			if (afs_log_status.w_termsig != 0) {
				lreply(230, "[ %s: signal %d ]",
					"AFS authentication failed",
					afs_log_status.w_termsig);
				break;
			}
			switch (afs_log_status.w_retcode) {
			case 0:
				lreply(230,
				       "Authenticated to local fileserver");
				break;
			case 1:
				lreply(230, "[ %s: uid not in passwd ]",
					"AFS authentication failed");
				break;
			case 2:
				lreply(230, "[ %s: usage error ]",
					"AFS authentication failed");
				break;
			case 3:
				lreply(230, "[ %s: user not in passwd ]",
					"AFS authentication failed");
				break;
			case 4:
				lreply(230, "[ %s: InitRPC failure ]",
					"AFS authentication failed");
				break;
			case 5:
				lreply(230, "[ %s: invalid login ]",
					"AFS authentication failed");
				break;
			case 6:
				lreply(230, "[ %s: Cache Manager ]",
					"AFS authentication failed");
				break;
			default:
				lreply(230, "[ %s: exit %d ]",
					"AFS authentication failed",
					afs_log_status.w_retcode);
				break;
			}
			break;
		case AFSERR_NOPASS: /* no password, no authentication */
			lreply(230, "[ %s: no password ]",
				"AFS authentication failed");
			break;
		case AFSERR_SIGUSED: /* signals in use, no authentication */
			lreply(230, "[ %s: signals in use ]",
				"AFS authentication failed");
			break;
		case AFSERR_PIPE: /* pipe syscall, no authentication */
			lreply(230, "[ %s: pipe: %s ]",
				"AFS authentication failed",
				errmsg(afs_log_status.w_status));
			break;
		case AFSERR_FORK: /* fork syscall, no authentication */
			lreply(230, "[ %s: fork: %s ]",
				"AFS authentication failed",
				errmsg(afs_log_status.w_status));
			break;
		case AFSERR_SETPAG: /* setpag syscall, no authentication */
			lreply(230, "[ %s: setpag: %s ]",
				"AFS authentication failed",
				errmsg(afs_log_status.w_status));
			break;
		case AFSERR_TIMEOUT: /* timed out log, maybe authenticated */
			lreply(230, "[ AFS authentication timed out ]");
			break;
		default:
			lreply(230, "[ %s: unknown errcode %d ]",
				"AFS authentication failed",
				afs_log_errcode);
			break;
		}
	}
#endif 	/* AFS */

	if (setegid(pw->pw_gid) < 0) {
		reply(421, "setegid() call failed - %s",errmsg(-1));
#ifndef VACFASTFTPD
		dologout(1);
#endif
	}
	if (setgroups(*grlist, grlist+1) < 0) {
		reply(421, "setgroups() call failed - %s",errmsg(-1));
#ifndef VACFASTFTPD
		dologout(1);
#endif
	}
	if (chdir(pw->pw_dir) < 0)
		chdir("/");

	abspath(0, 0);
	if (seteuid(pw->pw_uid) < 0) {
		reply(421, "seteuid() call failed - %s", errmsg(-1));
#ifndef VACFASTFTPD
		dologout(1);
#endif
	}

	logged_in = 1;

	/*
	 * give real users lots more time. Some people (and systems) 
	 * use ftp every so many minutes to store files away. In these
	 * cases, 2 minutes is too short.
	 */
	if (!guest && strcmp(userid, ANONUSER) != 0)
		timeout = 30*60;	/* half an hour */
	/*
	 * Save everything so globbing doesn't
	 * clobber the fields.
	 */
	save = *pw;
	save.pw_name = savestr(pw->pw_name);
	save.pw_passwd = savestr(pw->pw_passwd);
	save.pw_comment = savestr(pw->pw_comment);
	save.pw_gecos = savestr(pw->pw_gecos);
	save.pw_dir = savestr(pw->pw_dir);
	save.pw_shell = savestr(pw->pw_shell);
	pw = &save;
	home = pw->pw_dir;		/* home dir for globbing */
	return 1;
}

oklocal()
{
	char *p, buf[BUFSIZ];
	FILE *fp;
	int len;

	/* see if anon user exists */
	if (getpwnam(ANONUSER) == NULL)
		return(0);

	if ((fp = fopen("/usr/cs/lib/hosts/local-hosts", "r")) == NULL)
		return(0);
	len = strlen(rhostaddr) + 1;
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		if (p = index(buf, '\n'))
			*p = '\0';
		if (index("#;:",*buf))	/* allow comments */
			continue;
		if (bcmp(buf, rhostaddr, len) != 0)
			continue;
		fclose(fp);
		return(1);
	}
	fclose(fp);
	return(0);
}

freelogin()
{
	userid = ANONUSER;
	groupid = NULL;
#ifndef VACFASTFTPD
	dologin(NULL);
#endif
}

/*
 * don't allow ".." or commands in filename components.
 */
chkacc(arg)
char *arg;
{
	register char *p;
	register int sep;

	/* no commands */

	if (*arg == '|')
		return (1);
	/* no /.. sequences */
	sep = 1;
	for (p = arg; *p; p++) {
		if (sep && p[0] == '.' && p[1] == '.')
			return (1);
		sep = (p[0] == '/');
	}


	return (0);
}

#include <varargs.h>

/*VARARGS2*/
reply(va_alist)
	va_dcl
{
	int n;
	char *fmt;
	FILE _strbuf;
	char buf[BUFSIZ];
	va_list ap;
	int l;

	va_start(ap);
	n = va_arg(ap, int);
	fmt = va_arg(ap, char *);
	l = snprintf(buf, sizeof(buf), "%d ", n);
	l += vsnprintf(&buf[l], sizeof(buf)-l, fmt, ap);
	va_end(ap);
	fputs(buf, stdout);
	fputs("\r\n", stdout);
	fflush(stdout);
	if (logging)
		syslog(LOG_INFO, "%s", buf);
}

/*VARARGS2*/
lreply(va_alist)
	va_dcl
{
	int n;
	char *fmt;
	FILE _strbuf;
	char buf[BUFSIZ];
	va_list ap;
	int l;

	va_start(ap);
	n = va_arg(ap, int);
	fmt = va_arg(ap, char *);
	l = snprintf(buf, sizeof(buf), "%d-", n);
	l += vsnprintf(&buf[l], sizeof(buf)-l, fmt, ap);
	va_end(ap);
	fputs(buf, stdout);
	fputs("\r\n", stdout);
	fflush(stdout);
	if (logging)
		syslog(LOG_INFO, "%s", buf);
}
#endif	CMUCS
