/*
 * Copyright (c) 1980 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1980 Regents of the University of California.\n\
 All rights reserved.\n";
#endif not lint

#ifndef lint
static char sccsid[] = "@(#)mount.c	5.2 (Berkeley) 11/21/85";
#endif not lint

/*
 * mount
 **********************************************************************
 * HISTORY
 * $Log:	mount.c,v $
 * Revision 1.3  89/08/30  16:55:26  moore
 * 	Changed the 4.1 mtab conversion code to assume that the mtab
 * 	structure is 4.2 if you have a multiple of 16 file systems.
 * 
 * 	Dale Moore
 * 	[89/08/30  16:53:42  moore]
 * 
 * Revision 1.2  89/06/12  15:58:44  moore
 * 	Modified to use either 4.1 or 4.2 format of /etc/mtab.
 * 	[89/06/12  15:58:33  moore]
 * 
 * 06-Nov-80  Mike Accetta (mja) at Carnegie-Mellon University
 *	Changed to allow operator to mount file systems.
 *
 **********************************************************************
 */
#include <sys/param.h>

#include <stdio.h>
#include <fstab.h>
#include <mtab.h>
#include <errno.h>

#undef CMUCS

#define	DNMAX	(sizeof (mtab[0].m_dname) - 1)
#define	PNMAX	(sizeof (mtab[0].m_path) - 1)

struct	mtab mtab[NMOUNT];

int	all;
int	ro;
int	fake;
int	verbose;
char	*index(), *rindex();

main(argc, argv)
	int argc;
	char **argv;
{
	register struct mtab *mp;
	register char *np;
	int mf;
	char *type = FSTAB_RW;

#ifdef	CMUCS
	chkopr();
#endif
	mf = open("/etc/mtab", 0);
	if (mf >= 0) {
#if	CMUCS
		int l;

		/*
		 *  The mtab structure format was changed between 4.1 and 4.2
		 *  to include a trailing m_type field.  To facilitate
		 *  conversion to the new format, we detect the old format here
		 *  and convert internally to the new before using any of the
		 *  mtab data.  The old format had 64 bytes per record and the
		 *  new has 68 bytes per record which means the check below
		 *  will fail when non-zero multiples of 17 file systems are
		 *  mounted.
		 */
		l = read(mf, (char *)mtab, sizeof (mtab));
		if ((l > 0) && ((l%sizeof(struct mtab)) != 0))
		{
		    for (l=NMOUNT; l > 0; l--)
		    {
			char *cpt, *cpf; 
			int k;

			cpt = (char *)&mtab[l];
			cpf = cpt-(l*sizeof(mtab[0].m_type));
			cpt -= sizeof(mtab[0].m_type);
			for (k=sizeof(mtab[0])-sizeof(mtab[0].m_type); k-->0;)
			    *--cpt = *--cpf;
			bzero((char *)mtab[l].m_type, sizeof(mtab[0].m_type));
		    }
		}
#else	CMUCS
		(void) read(mf, (char *)mtab, sizeof (mtab));
#endif	CMUCS
	}
	if (argc == 1) {
		for (mp = mtab; mp < &mtab[NMOUNT]; mp++)
			if (mp->m_path[0] != '\0')
				prmtab(mp);
		exit(0);
	}
top:
	if (argc > 1) {
		if (!strcmp(argv[1], "-a")) {
			all++;
			argc--, argv++;
			goto top;
		}
		if (!strcmp(argv[1], "-r")) {
			type = FSTAB_RO;
			argc--, argv++;
			goto top;
		}
		if (!strcmp(argv[1], "-f")) {
			fake++;
			argc--, argv++;
			goto top;
		}
		if (!strcmp(argv[1], "-v")) {
			verbose++;
			argc--, argv++;
			goto top;
		}
	}
	if (all) {
		struct fstab *fsp;

		if (argc > 1)
			goto argcnt;
		close(2); dup(1);
		if (setfsent() == 0)
			perror(FSTAB), exit(1);
		while ((fsp = getfsent()) != 0) {
			if (strcmp(fsp->fs_file, "/") == 0)
				continue;
			if (strcmp(fsp->fs_type, FSTAB_RO) &&
			    strcmp(fsp->fs_type, FSTAB_RW) &&
			    strcmp(fsp->fs_type, FSTAB_RQ))
				continue;
			mountfs(fsp->fs_spec, fsp->fs_file, fsp->fs_type);
		}
		exit(0);
	}
	if (argc == 2) {
		struct fstab *fs;

		if (setfsent() == 0)
			perror(FSTAB), exit(1);
		fs = getfsfile(argv[1]);
		if (fs == NULL)
			goto argcnt;
		if (strcmp(fs->fs_type, FSTAB_RO) &&
		    strcmp(fs->fs_type, FSTAB_RW) &&
		    strcmp(fs->fs_type, FSTAB_RQ))
			goto argcnt;
		mountfs(fs->fs_spec, fs->fs_file, fs->fs_type);
		exit(0);
	}
	if (argc != 3) {
argcnt:
		fprintf(stderr,
    "usage: mount [ -a ] [ -r ] [ -f ] [ -v ] [ special dir ] [ dir ]\n");
		exit(1);
	}
	mountfs(argv[1], argv[2], type);
}

prmtab(mp)
	register struct mtab *mp;
{

	printf("%s on %s", mp->m_dname, mp->m_path);
	if (strcmp(mp->m_type, FSTAB_RO) == 0)
		printf("\t(read-only)");
	if (strcmp(mp->m_type, FSTAB_RQ) == 0)
		printf("\t(with quotas)");
	putchar('\n');
}

mountfs(spec, name, type)
	char *spec, *name, *type;
{
	register char *np;
	register struct mtab *mp;
	int mf;

	if (!fake) {
		if (mount(spec, name, strcmp(type, FSTAB_RO) == 0) < 0) {
			extern int errno;
			char *cp;

			fprintf(stderr, "%s on ", spec);
			switch (errno) {

			case EMFILE:
				cp = "Mount table full";
				break;

			case EINVAL:
				cp = "Bogus super block";
				break;

			default:
				perror(name);
				return;
			}
			fprintf(stderr, "%s: %s\n", name, cp);
			return;
		}
		/* we don't do quotas.... */
		if (strcmp(type, FSTAB_RQ) == 0)
			type = FSTAB_RW;
	}
	np = index(spec, '\0');
	while (*--np == '/')
		*np = '\0';
	np = rindex(spec, '/');
	if (np) {
		*np++ = '\0';
		spec = np;
	}
	for (mp = mtab; mp < &mtab[NMOUNT]; mp++)
		if (strcmp(mp->m_dname, spec) == 0)
			goto replace;
	for (mp = mtab; mp < &mtab[NMOUNT]; mp++)
		if (mp->m_path[0] == '\0')
			goto replace;
	return;
replace:
	strncpy(mp->m_dname, spec, DNMAX);
	mp->m_dname[DNMAX] = '\0';
	strncpy(mp->m_path, name, PNMAX);
	mp->m_path[PNMAX] = '\0';
	strcpy(mp->m_type, type);
	if (verbose)
		prmtab(mp);
	mp = mtab + NMOUNT - 1;
	while (mp > mtab && mp->m_path[0] == '\0')
		--mp;
	mf = creat("/etc/mtab", 0644);
	write(mf, (char *)mtab, (mp - mtab + 1) * sizeof (struct mtab));
	close(mf);
	return;
}
