#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define ANY_PROTOCOL ((char *) 0)

#define KILO (1024)
#define MEGA (KILO*KILO)

#define MICRO (1000*1000)

#define CLOCKDIFF(startp, endp) ((((endp)->tv_sec - (startp)->tv_sec) * MICRO) + ((endp)->tv_usec - (startp)->tv_usec))

#define MIN(x,y) (((x) < (y)) ? (x) : (y))

extern char *malloc();
main(argc, argv)

int argc;
char **argv;
{
	char *myname;
	char *hostname;
	int sock;
	char *buffer;
	int buffersize, nbuffers, bytes, nbytes;

	struct timeval starttime;
	struct timeval endtime;

	struct servent *service;
	struct protoent *proto;
	struct hostent *host;
	struct sockaddr_in address;

	myname = argv[0];
	
	if (argc != 4) {
		printf("usage: %s hostname buffersize #buffers\n", myname);
		exit(1);
	}

	hostname = argv[1];

	if (((buffersize = atoi(argv[2])) < 10) || (buffersize > (10*MEGA))) {
		printf("%s: silly buffersize %d.\n", myname, buffersize);
		exit(2);
	}

	if (((nbuffers = atoi(argv[3])) < 1) || (nbuffers > (MEGA))) {
		printf("%s: silly nbuffers %d.\n", myname, nbuffers);
		exit(3);
	}

	if (!(buffer = malloc(buffersize))) {
		printf("%s: couldn't malloc %d bytes.\n", myname, buffersize);
		exit(4);
	}

	if (!(service = getservbyname("daytime", ANY_PROTOCOL))) {
		printf("daytime: unknown service\n");
		exit(5);
	}
	printf("daytime is %s port %d.\n", service->s_proto, service->s_port);

	if (!(proto = getprotobyname(service->s_proto))) {
		printf("%s: unknown protocol\n", service->s_proto);
		exit(6);
	}
	printf("%s is protocol %d.\n", service->s_proto, proto->p_proto);

	if (!(host = gethostbyname(argv[1]))) {
		printf("%s: unknown host %s\n", myname, argv[1]);
		exit(7);
	}
	printf("%s is a good host.\n", argv[1]);

	bzero(address, sizeof (address));
	address.sin_family = host->h_addrtype;
	address.sin_port = service->s_port;
	bcopy(host->h_addr, &address.sin_addr, host->h_length);

	if ((sock = socket(PF_INET, SOCK_STREAM, proto->p_proto)) < 0) {
		perror("socket");
		exit(8);
	}
	if (connect(sock, &address, sizeof (address))) {
		perror("connect");
		exit(9);
	}

	gettimeofday(&starttime, 0);

	if ((nbytes = nbuffers * buffersize) < buffersize) {
		/* rough check for overflow */
		printf("%s: %d * %d = %d:  integer overflow?\n", nbuffers, buffersize, nbytes);
		exit(10);
	}

	bytes = nbytes;
	while (bytes > 0) {
		int readsize;
		if ((readsize = read(sock, buffer, MIN(buffersize, bytes))) < 0) {
			perror("read");
		} else {
			write(1, buffer, readsize);
			bytes -= readsize;
		}
	}

	gettimeofday(&endtime, 0);

	printf("microseconds: %ld.\n", CLOCKDIFF(&starttime, &endtime));

	printf("Throughput: %lg kbytes/sec.\n", 1000. * (double) ((double) nbytes / (double) CLOCKDIFF(&starttime, &endtime)));

	exit(0);
}
