/*
 **********************************************************************
 * HISTORY
 * $Log:	df.c,v $
 * Revision 1.5  91/01/06  17:13:53  baird
 * 	Added code to process a root entry for compatibility with old /etc/mtab
 * 	files that do not have one.  If the first entry in mtab is not root, it
 * 	uses the first entry from fstab.  Also changed to not core dump when -t
 * 	is not given any arguments.
 * 	[91/01/06            baird]
 * 
 * Revision 1.4  91/01/06  00:17:07  baird
 * 	Changed type of oldsig to match with latest definition of signal.
 * 	[91/01/05            baird]
 * 
 * 	Created from MTXINU revision 2.2
 * 	Added code to do old style df (by reading sblock) if the kernel
 * 	does not support the statfs call.  Changed mount types to be
 * 	mount options.
 * 	[90/10/23            baird]
 * 
 * Revision 2.2  89/10/06  13:09:45  bww
 * 	NFS capable version.
 * 	[89/10/06  13:09:24  bww]
 * 
 **********************************************************************
 */
/*
 * Copyright (c) 1980 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1980 Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)df.c	5.3 (Berkeley) 1/10/88";
#endif /* not lint */

#include <sys/param.h>
#include <errno.h>
#include <sys/fs.h>
#include <sys/vfs.h>
#include <sys/stat.h>
#if CMUCS
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/syscall.h>
#endif /* CMUCS */
#include <stdio.h>
#include <mntent.h>

extern int errno;

char	*mpath();
int	iflag;
int	aflag;
int	type;
char	*typestr;
#if CMUCS
int didroot;
struct mntent *root;		/* a mntent for for "/" from fstab since */
				/* old cmucs mtab do not an entry for root */
struct mntent * getroot();
#endif /* CMUCS */

union {
	struct fs iu_fs;
	char dummy[SBSIZE];
} sb;
#define sblock sb.iu_fs

daddr_t	alloc();
char	*strcpy();

/*
 * This structure is used to build a list of mntent structures
 * in reverse order from /etc/mtab.
 */
struct mntlist {
	struct mntent *mntl_mnt;
	struct mntlist *mntl_next;
};

struct mntlist *mkmntlist();
struct mntent *mntdup();

main(argc, argv)
	int argc;
	char **argv;
{
	register struct mntent *mnt;
	register int i;

	argv++;
	while (argc > 1 && argv[0][0]=='-') {
		switch (argv[0][1]) {

		case 'i':
			iflag++;
			break;

		case 't':
			type++;
#if CMUCS
			if (argv[1] && (strcmp(argv[1], OMNTTYPE_42) == 0
			    || strcmp(argv[1], OMNTTYPE_43) == 0))
				typestr = MNTTYPE_UFS;
			else
				typestr = (argv[1] ? argv[1] : MNTTYPE_UFS);
#else /* CMUCS */
			if (strcmp(argv[1], MNTTYPE_42) == 0
			    || strcmp(argv[1], MNTTYPE_43) == 0)
				typestr = MNTTYPE_UFS;
			else
				typestr = argv[1];
#endif /* CMUCS */
			argv++;
			argc--;
			break;

		case 'a':
			aflag++;
			break;

		default:
			usage();
		}
		argc--, argv++;
	}
	if (argc > 1 && type) {
		usage();
	}
	sync();
	printf("Filesystem            kbytes    used   avail capacity");
	if (iflag)
		printf(" iused   ifree  %%iused");
	printf("  Mounted on\n");
#if CMUCS
	root = getroot();
#endif /* CMUCS */
	if (argc <= 1) {
		FILE *mtabp;

		if ((mtabp = setmntent(MOUNTED, "r")) == 0) {
			perror(MOUNTED);
			exit(1);
		}
		while (mnt = getmntent(mtabp)) {
#if CMUCS
			if (!didroot) {
				doroot(mnt, typestr);
				didroot++;
			}
			if (strcmp(mnt->mnt_type, MNTOPT_IGNORE) == 0 ||
			    strcmp(mnt->mnt_type, MNTOPT_SWAP) == 0)
#else /* CMUCS */
			if (strcmp(mnt->mnt_type, MNTTYPE_IGNORE) == 0 ||
			    strcmp(mnt->mnt_type, MNTTYPE_SWAP) == 0)
#endif /* CMUCS */
				continue;
			if (type && strcmp(typestr, mnt->mnt_type)) {
				continue;
			}
			dfreemnt(mnt->mnt_dir, mnt);
		}
		(void) endmntent(mtabp);
	} else {
		int num = argc;
		struct mntlist *mntl;

		aflag++;
		/*
		 * Reverse the list and start comparing.
		 */
		for (mntl = mkmntlist(); mntl != NULL && num;
		    mntl = mntl->mntl_next) {
			struct stat dirstat, filestat;

			mnt = mntl->mntl_mnt;
			if (stat(mnt->mnt_dir, &dirstat) < 0)
				continue;

			for (i = 0; i < argc; i++) {
			    if (argv[i] == NULL)
				continue;
			    if (stat(argv[i], &filestat) < 0) {
				(void) fprintf(stderr, "df: ");
				perror(argv[i]);
				argv[i] = NULL;
				--num;
			    } else {
				if ((filestat.st_mode & S_IFMT) == S_IFBLK ||
				    (filestat.st_mode & S_IFMT) == S_IFCHR) {
					char *cp;

					cp = mpath(argv[i]);
					if (*cp == '\0') {
						dfreedev(argv[i]);
						argv[i] = NULL;
						--num;
						continue;
					} else if (stat(cp, &filestat) < 0) {
						(void) fprintf(stderr, "df: ");
						perror(argv[i]);
						argv[i] = NULL;
						--num;
						continue;
					}
				}
#if CMUCS
				if (strcmp(mnt->mnt_type,MNTOPT_IGNORE) == 0 ||
				    strcmp(mnt->mnt_type, MNTOPT_SWAP) == 0)
#else /* CMUCS */
				if (strcmp(mnt->mnt_type,MNTTYPE_IGNORE) == 0 ||
				    strcmp(mnt->mnt_type, MNTTYPE_SWAP) == 0)
#endif /* CMUCS */
					continue;
				if (filestat.st_dev == dirstat.st_dev) {
					dfreemnt(mnt->mnt_dir, mnt);
					argv[i] = NULL;
					--num;
				}
			    }
			}
		}
		if (num) {
			for (i = 0; i < argc; i++) {
				if (argv[i] == NULL)
					continue;
				(void) fprintf(stderr,"No mount point for %s\n",
						argv[i]);
			}
		}
	}
	exit(0);
	/*NOTREACHED*/
}

dfreedev(file)
	char *file;
{
	long totalblks, availblks, avail, free, used;
	int fi;

	fi = open(file, 0);
	if (fi < 0) {
		perror(file);
		return;
	}
	if (bread(file, fi, SBLOCK, (char *)&sblock, SBSIZE) == 0) {
		(void) close(fi);
		return;
	}
	if (sblock.fs_magic != FS_MAGIC) {
		(void) fprintf(stderr, "df: %s: not a UNIX filesystem\n", file);
		(void) close(fi);
		return;
	}
	(void) printf("%-20.20s", file);
	totalblks = sblock.fs_dsize;
	free = sblock.fs_cstotal.cs_nbfree * sblock.fs_frag +
	    sblock.fs_cstotal.cs_nffree;
	used = totalblks - free;
	availblks = totalblks * (100 - sblock.fs_minfree) / 100;
	avail = availblks > used ? availblks - used : 0;
	printf("%8d%8d%8d",
	    totalblks * sblock.fs_fsize / 1024,
	    used * sblock.fs_fsize / 1024,
	    avail * sblock.fs_fsize / 1024);
	printf("%6.0f%%",
	    availblks==0? 0.0: (double)used/(double)availblks * 100.0);
	if (iflag) {
		int inodes = sblock.fs_ncg * sblock.fs_ipg;
		used = inodes - sblock.fs_cstotal.cs_nifree;
		printf("%8ld%8ld%6.0f%% ", used, sblock.fs_cstotal.cs_nifree,
		    inodes == 0 ? 0.0 : (double)used / (double)inodes * 100.0);
	} else
		printf("  ");
	printf("  %s\n", mpath(file));
	(void) close(fi);
}

dfreemnt(file, mnt)
	char *file;
	struct mntent *mnt;
{
	struct statfs fs;
	long totalblks, avail, free, used, reserved;
#if CMUCS
	void (*oldsig)();
	int callstatus, error;
	extern int errno;

	oldsig = signal(SIGSYS, SIG_IGN);
	callstatus = syscall(SYS_statfs, file, &fs);
        error = errno;
        (void) signal(SIGSYS, oldsig);
        errno = error;

	if (callstatus < 0) {
		if (errno == EINVAL) {
			olddfree(file, mnt);
			return;
		} else {
			(void) fprintf(stderr, "df: ");
			perror(file);
			return;
		}
	}
#else /* CMUCS */
	if (statfs(file, &fs) < 0) {
		(void) fprintf(stderr, "df: ");
		perror(file);
		return;
	}
#endif /* CMUCS */

	if (!aflag && fs.f_blocks == 0) {
		return;
	}
	if (strlen(mnt->mnt_fsname) > 20) {
		printf("%s\n", mnt->mnt_fsname);
		printf("                    ");
	} else {
		printf("%-20.20s", mnt->mnt_fsname);
	}
	totalblks = fs.f_blocks;
	free = fs.f_bfree;
	used = totalblks - free;
	avail = fs.f_bavail;
	reserved = free - avail;
	if (avail < 0)
		avail = 0;
	printf("%8d%8d%8d", totalblks * fs.f_bsize / 1024,
	    used * fs.f_bsize / 1024, avail * fs.f_bsize / 1024);
	totalblks -= reserved;
	printf("%6.0f%%",
	    totalblks==0? 0.0: (double)used/(double)totalblks * 100.0);
	if (iflag) {
		long files, used;

		files = fs.f_files;
		used = files - fs.f_ffree;
		printf("%8ld%8ld%6.0f%% ", used, fs.f_ffree,
		    files == 0? 0.0: (double)used / (double)files * 100.0);
	} else
		printf("  ");
	printf("  %s\n", mnt->mnt_dir);
}

/*
 * Given a name like /dev/rrp0h, returns the mounted path, like /usr.
 */
char *
mpath(file)
	char *file;
{
	FILE *mntp;
	register struct mntent *mnt;

#if CMUCS
	if (root && (strcmp(file, root->mnt_fsname) == 0))
		return (root->mnt_dir);
#endif /* CMUCS */

	if ((mntp = setmntent(MOUNTED, "r")) == 0) {
		(void) fprintf(stderr, "df: ");
		perror(MOUNTED);
		exit(1);
	}

	while ((mnt = getmntent(mntp)) != 0) {
		if (strcmp(file, mnt->mnt_fsname) == 0) {
			endmntent(mntp);
			return (mnt->mnt_dir);
		}
	}
	endmntent(mntp);
	return "";
}

long lseek();

bread(file, fi, bno, buf, cnt)
	char *file;
	int fi;
	daddr_t bno;
	char *buf;
	int cnt;
{
	int n;

	(void) lseek(fi, (long)(bno * DEV_BSIZE), 0);
	if ((n=read(fi, buf, cnt)) != cnt) {
		/* probably a dismounted disk if errno == EIO */
		if (errno != EIO) {
			printf("\nread(bno = %ld) == %d", bno, n);
			fflush(stdout);
			perror("");
		}
		return (0);
	}
	return (1);
}

char *
xmalloc(size)
	unsigned int size;
{
	char *ret;

	if ((ret = (char *)malloc(size)) == NULL) {
		(void) fprintf(stderr, "df: ran out of memory!\n");
		exit(1);
	}
	return (ret);
}

struct mntent *
mntdup(mnt)
	register struct mntent *mnt;
{
	register struct mntent *new;

	new = (struct mntent *)xmalloc(sizeof(*new));

#if CMUCS
	if (!mnt->mnt_fsname) mnt->mnt_fsname = "";
	new->mnt_fsname = (char *)xmalloc(strlen(mnt->mnt_fsname) + 1);
	strcpy(new->mnt_fsname, mnt->mnt_fsname);

	if (!mnt->mnt_dir) mnt->mnt_dir = "";
	new->mnt_dir = (char *)xmalloc(strlen(mnt->mnt_dir) + 1);
	strcpy(new->mnt_dir, mnt->mnt_dir);

	if (!mnt->mnt_type) mnt->mnt_type = "";
	new->mnt_type = (char *)xmalloc(strlen(mnt->mnt_type) + 1);
	strcpy(new->mnt_type, mnt->mnt_type);

	if (!mnt->mnt_opts) mnt->mnt_opts = "";
	new->mnt_opts = (char *)xmalloc(strlen(mnt->mnt_opts) + 1);
	strcpy(new->mnt_opts, mnt->mnt_opts);

	if (!mnt->mnt_opts2) mnt->mnt_opts2 = "";
	new->mnt_opts2 = (char *)xmalloc(strlen(mnt->mnt_opts2) + 1);
	strcpy(new->mnt_opts2, mnt->mnt_opts2);
#else /* CMUCS */
	new->mnt_fsname = (char *)xmalloc(strlen(mnt->mnt_fsname) + 1);
	strcpy(new->mnt_fsname, mnt->mnt_fsname);

	new->mnt_dir = (char *)xmalloc(strlen(mnt->mnt_dir) + 1);
	strcpy(new->mnt_dir, mnt->mnt_dir);

	new->mnt_type = (char *)xmalloc(strlen(mnt->mnt_type) + 1);
	strcpy(new->mnt_type, mnt->mnt_type);

	new->mnt_opts = (char *)xmalloc(strlen(mnt->mnt_opts) + 1);
	strcpy(new->mnt_opts, mnt->mnt_opts);
#endif /* CMUCS */
	new->mnt_freq = mnt->mnt_freq;
	new->mnt_passno = mnt->mnt_passno;

	return (new);
}

struct mntlist *
mkmntlist()
{
	FILE *mounted;
	struct mntlist *mntl;
	struct mntlist *mntst = NULL;
	struct mntent *mnt;

	if ((mounted = setmntent(MOUNTED, "r"))== NULL) {
		(void) fprintf(stderr, "df : ") ;
		perror(MOUNTED);
		exit(1);
	}
	while ((mnt = getmntent(mounted)) != NULL) {
#if CMUCS
		if (!didroot) {
			if (root && strcmp(root->mnt_dir, mnt->mnt_dir)) {
				mntl = (struct mntlist *) xmalloc(sizeof(*mntl));
				mntl->mntl_mnt = mntdup(root);
				mntl->mntl_next = mntst;
				mntst = mntl;
			}
			didroot++;
		}
#endif /* CMUCS */
		mntl = (struct mntlist *) xmalloc(sizeof(*mntl));
		mntl->mntl_mnt = mntdup(mnt);
		mntl->mntl_next = mntst;
		mntst = mntl;
	}
	(void) endmntent(mounted);
	return(mntst);
}

usage()
{

	fprintf(stderr, "usage: df [ -i ] [-t type | file... ]\n");
	exit(0);
}

#if CMUCS

/*
 * Olddfree is used on older kernels that do not support the
 * statfs system call (primarily CS5 kernels).  This routine just
 * read the old sblock and computes df info.  It should go away when
 * CS7 is everywhere.  (Or if mntent library was reentrant this could
 * be replaced by dfreedev(mnt->mnt_fsname).)
 */

olddfree(file, mnt)
	char *file;		/* Not used */
	struct mntent *mnt;
{
	long totalblks, availblks, avail, free, used;
	int fi;

	fi = open(mnt->mnt_fsname, 0);
	if (fi < 0) {
		perror(mnt->mnt_fsname);
		return;
	}
	if (bread(mnt->mnt_fsname, fi, SBLOCK, (char *)&sblock, SBSIZE) == 0) {
		(void) close(fi);
		return;
	}
	if (sblock.fs_magic != FS_MAGIC) {
		(void) fprintf(stderr, "df: %s: not a UNIX filesystem\n", mnt->mnt_fsname);
		(void) close(fi);
		return;
	}
	(void) printf("%-20.20s", mnt->mnt_fsname);
	totalblks = sblock.fs_dsize;
	free = sblock.fs_cstotal.cs_nbfree * sblock.fs_frag +
	    sblock.fs_cstotal.cs_nffree;
	used = totalblks - free;
	availblks = totalblks * (100 - sblock.fs_minfree) / 100;
	avail = availblks > used ? availblks - used : 0;
	printf("%8d%8d%8d",
	    totalblks * sblock.fs_fsize / 1024,
	    used * sblock.fs_fsize / 1024,
	    avail * sblock.fs_fsize / 1024);
	printf("%6.0f%%",
	    availblks==0? 0.0: (double)used/(double)availblks * 100.0);
	if (iflag) {
		int inodes = sblock.fs_ncg * sblock.fs_ipg;
		used = inodes - sblock.fs_cstotal.cs_nifree;
		printf("%8ld%8ld%6.0f%% ", used, sblock.fs_cstotal.cs_nifree,
		    inodes == 0 ? 0.0 : (double)used / (double)inodes * 100.0);
	} else
		printf("  ");
	printf("  %s\n", mnt->mnt_dir);
	(void) close(fi);
}

struct mntent * 
getroot()
{
	FILE *fstabp;
	struct mntent *root;

	fstabp = setmntent(FSTAB, "r");
	root = getmntent(fstabp);
	endmntent(fstabp);
	if (root) {
		root = mntdup(root);
	}
	return (root);
}

doroot(mnt, typestr)
	struct mntent *mnt;
	char *typestr;
{
	if (!root || (typestr && strcmp(typestr, MNTTYPE_UFS)))
	  return;
	if (strcmp(mnt->mnt_dir, "/"))
	  dfreemnt("/", root);
}

#endif /* CMUCS */
