/*
 *  This is to the face program glue.  It hooks frame grabber
 *  and display code from Steve Schlick and iwarp Halftoning code
 *  from John Webb.
 *
 *  Communication with the iwarp is over the named pipes stdin0 and stdout0.
 *
 *  Communication with the process on the machine with the 
 *  cammera is via pipes to the child doing an rsh.
 *
 *  It prompts for a user ID and when this is entered, does the
 *  frame grab and sends the data down a pipe stdin0 to the 
 *  iwarp.   It gets the halftoned data back on pipe stdout0.
 *  It then takes this and saves it to a file using the user ID
 *  as part of the name.
 *
 */

#include <stdio.h>
#include <strings.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>


#define BASENAME "/afs/cs/user/vac/face/"    

/* #define BASENAME "image." */

#define RAWSIZE 512*480 
#define WARPEDSIZE 6+(600*480/8)

#define WORLDRW  S_IREAD | S_IWRITE | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH
#define WORLDR  S_IREAD | S_IRGRP | S_IROTH 

struct sockaddr_un {
   short    sun_family;
   u_short  sun_path[108];
};

FILE *toIwarp;

int  FDfromIwarp;

int FDtoGrabber, FDfromGrabber;         /* file descriptors for grabber */

int source;         /* 'f' means file otherwise grabber */

fail(s)
char *s;
{
     perror(s);
     exit(-1);
}

IwarpInit() 
{
    struct sockaddr_un namein, nameout;
    int namelen, error, oldumask;
    mode_t mymode;
    char   stdin0str[30];
    char   stdout0str[30];

    printf("Starting IwarpInit\n");
    fflush(stdout); 

    strcpy(stdin0str, "stdin0");
    strcpy(stdout0str, "stdout0");

    unlink(stdin0str);    /* ) fail("Can not unlink stdin0"); */
    unlink(stdout0str);   /* ) fail("Can not unlink stdout0"); */

    oldumask=umask(0);

    mymode=WORLDRW;
    if (mknod(stdin0str, S_IFIFO|0666, 0) !=0) fail ("Can not make stdin0");
    if (mknod(stdout0str, S_IFIFO|0666, 0) !=0) fail ("Can not make stdout0");

    printf("Now for opens \n");
    fflush(stdout); 

    toIwarp=fopen(stdin0str, "w+");
    printf("Back from open for write\n");
    fflush(stdout); 
    if (toIwarp==NULL) fail ("Can not open stdin0");

    FDfromIwarp=open(stdout0str, O_RDONLY | O_NDELAY); 
    printf("Back from open for read\n");
    fflush(stdout); 
    if (FDfromIwarp<0) fail ("Can not open stdout0");

    printf("Done with IwarpInit\n");
    fflush(stdout); 

    umask(oldumask);
}

GrabberInit()
{
    int pid, i, c;
    int toremote[2], fromremote[2];

    printf("Starting IwarpInit\n");
    fflush(stdout); 

    if (pipe(toremote) !=0) fail("Can not open pipe toremote");
    if (pipe(fromremote) !=0) fail("Can not open pipe fromremote");

    pid = fork();
    if (pid == 0) {
        if (dup2(toremote[0], 0) <0) fail("Child dup2 of toremote0");
        if (dup2(fromremote[1], 1) <0) fail("Child dup2 of fromrmote1");
        execl("/usr/ucb/rsh", "rsh", "bual", "/usr/vac/face/remotegrab",0);
        fail("Execl did not work");
    }

    FDtoGrabber=toremote[1];
    FDfromGrabber=fromremote[0];

    /* Grabber=popen("/usr/ucb/rsh bual face3/remotegrab", "wr"); */
    /* if (Grabber == NULL) fail("Can not start grabber");           */

    printf("Done with GrabberInit\n");
    fflush(stdout); 
}

CharToGrabber(c)
char c;
{
    int i;
    char s[10];

    if (source=='f') return;

    s[1]=0;
    s[0]=c;
    if (write(FDtoGrabber, s, 1) == EOF) fail("Can not write %s to Grabber",s);

    /* 4 silly lines of code.  I really really want a   flush(FDtoGrabber)   */
    s[0]=' ';
    for (i=0; i++ ; i<8000) {
       if (write(FDtoGrabber, s, 1) == EOF) fail("Can not write to Grabber");
    }
}
   
int getonechar()
{
    char s[100];
    int result;

    gets(s);
    result=s[0];
    if (isupper(result)) {
        result=tolower(result);
    }
    return(result);
}
 
main()
{
    char  userid[50], thisname[132], tmp;
    char  basename[132], command[200];
    int good, yesno, lastimage;

    IwarpInit();         /* pipes for talking to iwarp */

    printf("Read from (f)iles or (g)rabber? \n");
    source=getonechar();

    printf("Save last image? \n");
    lastimage=getonechar();

    if (source != 'f') {
       GrabberInit();       /* pipe for talking to frame grabber */
    } 
    strcpy(basename, BASENAME);

    printf("Entering main loop. \n");
    fflush(stdout); 

    while (1) {
        if (lastimage != 'y') CharToGrabber('l');

        printf("\n  Enter UserID: ");
	gets(&userid[0]);

	if (strcmp(userid, "quit") == 0) {
	    break;
	}

	printf("Funny face [n]? ", yesno);
	yesno=getonechar();
	if ((yesno == 'y') || (yesno == 'Y'))
	    sprintf(thisname, "%s%s.funny", basename, userid);
	else
	    sprintf(thisname, "%s%s", basename, userid);
	
        good='n';
        while (good=='n') {
           if (lastimage != 'y') CharToGrabber('l');
	   printf("Press return when ready to digitize...");
	   tmp=getonechar();
           if (lastimage != 'y') CharToGrabber('f');
           lastimage='n';
           printf("Is this a good picture?\n");
           good=getonechar();
           if (good != 'n') CharToGrabber('s');
        }

        if (source=='f') {
            FDfromGrabber=open(thisname, O_RDONLY);
            strcat(thisname, ".out");
            if (FDfromGrabber < 0) fail("Can not open raster file");
        }

        GrabAndSendToIwarp();

        if (source=='f') {
            close(FDfromGrabber);
        }

	printf("Iwarp is now doing halftone\n");

	if (ReadAndSaveImage(thisname) <0) fail("Can not save file");

        sprintf(command, "/afs/cs/user/bsy/bin/face -F %s", thisname);
        system(command);

        printf("[Hit return to clear display]\n");
        tmp=getonechar();
    }

    CharToGrabber('q');
}


/* Get RAWSIZE bytes from framegrabber and send to Iwarp */
GrabAndSendToIwarp()
{
    int copied, newbytes;
    char data[RAWSIZE];

    printf("Starting copy from frame grabber to Iwarp \n");
    copied=0;
    while (copied < RAWSIZE) {
        newbytes=read(FDfromGrabber, &data[0], RAWSIZE-copied);
        if (newbytes < 0) {
            fprintf(stderr, "Can not read from grabber. \n");
            exit(-1);
        }
        if (newbytes>0) {
            fwrite(&data[0], sizeof(char), newbytes, toIwarp);
            copied += newbytes;
        } else {
            sleep(1);
        }
    }
    printf("Done with copy from frame grabber to Iwarp \n");
}


/* Read WARPSIZED bytes from Iwarp and write to filename  */
int ReadAndSaveImage(filename)
char *filename;
{
    int fd, copied, newbytes, doneprintf;
    char data[WARPEDSIZE];

    fd = open(filename, O_WRONLY | O_CREAT, 0644);
    if (fd<0) return(-1);

    copied=0;
    newbytes=0;
    doneprintf=0;
    while (copied < WARPEDSIZE) {
        newbytes=read(FDfromIwarp, &data[0], WARPEDSIZE-copied);
        if (newbytes>0) {
           write(fd, &data[0], newbytes);
           copied += newbytes;
        } else {
           sleep(1);
        }

        if ((copied!=0) && (doneprintf==0)) {
           printf("Starting to copy from Iwarp into file %s\n",filename);
           doneprintf=1;
        }
    }

    close(fd);

    printf("Done saving bitmap to file \n");
    return(0);
}


