/*
 **********************************************************************
 * HISTORY
 * 02-Jan-88  Glenn Marcy (gm0w) at Carnegie-Mellon University
 *	Upgraded to 4.7.3 of BIND resolver module.  This changed
 *	dn_skip() to dn_skipname().
 *
 * 05-Nov-87  David VanRyzin (vanryzin) at Carnegie-Mellon University
 *	Added meaningful exit status.
 *
 * 20-Mar-86  Glenn Marcy (gm0w) at Carnegie-Mellon University
 *	Created.
 *
 **********************************************************************
 */
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
/* #include <netinet/in.h>  vac */
#include <netdb.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <libc.h>
#include <c.h>

int debug;
extern int h_trace;
extern int h_errno;
extern int rs_errno;

usage(progname)
char *progname;
{
    quit(-1, "usage: %s [ %s-name || %s-number ] ...\n", progname, progname, progname);
}

/* network number to ascii */
char *inet_nntoa(addr)
long addr;
{
    char *net;
    struct in_addr in;

    in.s_addr = addr;
    in.s_addr = ntohl((u_long)inet_netof(in));
    net = inet_ntoa(in);
    addr = ntohl((u_long)addr);
    if (IN_CLASSA(addr)) net += 6;
    else if (IN_CLASSB(addr)) net += 4;
    else if (IN_CLASSC(addr)) net += 2;
    return(net);
}

main(argc, argv)
int argc;
char **argv;
{
    char *progname, *prefix;
    int status = 0;

    debug = FALSE;
    _res.options &= ~RES_DEFNAMES;
    progname = rindex(argv[0], '/');
    if (progname++ == NULL)
	progname = argv[0];
    if (strncmp(progname, "host", 4) == 0) {
	progname[4] = '\0';
	prefix = "H";
    } else if (strncmp(progname, "net", 3) == 0) {
	progname[3] = '\0';
	prefix = "N";
    } else if (strncmp(progname, "service", 7) == 0) {
	progname[7] = '\0';
	if (argc < 2)
	    usage(progname);
	prefix = "S";
    } else if (strncmp(progname, "protocol", 8) == 0) {
	progname[8] = '\0';
	if (argc < 2)
	    usage(progname);
	prefix = "P";
    } else
	quit(-1, "%s: must be called host, net, service or protocol\n", 
	     progname);
    while (argc > 1 && argv[1][0] == '-') {
	if (strcmp(argv[1], "-d") == 0) {
	    _res.options |= RES_DEBUG;
	    debug++;
	} else
	    quit(-1, "%s: unknown switch '%s'\n", progname, argv[1]);
	argc--;
	argv++;
    }
    if (argc == 1) {
	char *p, hostname[BUFSIZ];

	if (gethostname(hostname, sizeof(hostname)-2) < 0 ||
	    hostname[0] == '\0')
	    quit(-1, "couldn't find current hostname\n");
	p = hostname + strlen(hostname);
	if (p[-1] != '.') {
	    *p++ = '.';
	    *p = '\0';
	}
	if (*prefix == 'N') {
	    struct hostent *hp;

	    if ((hp = gethostbyname(hostname)) == NULL)
		quit(-1, "couldn't find %s in the host table\n", hostname);
	    (void)strcpy(hostname, inet_nntoa(*(long *)hp->h_addr));
	}
	exit(printent(progname, prefix, hostname));
    }
    while (argc-- > 1) {
	int retval = printent(progname, prefix, argv[1]);
	if (retval)
		status = retval;
	argv++;
    }
    exit(status);
}

printent(progname, prefix, arg)
char *progname, *prefix, *arg;
{
    struct hostent *h;
    struct netent *n;
    struct servent *s;
    struct protoent *p;
    char *cpu, *os;
    char *eom, *cp, **pp;
    int first, cc;
    struct in_addr inet_makeaddr();
    long addr;
    int network;
    char packet[PACKETSZ];
    char answer[PACKETSZ];
    HEADER *hp;

    switch (*prefix) {
    case 'H':
	if (isdigit(*arg) && (addr = inet_addr(arg)) != -1)
	    h = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
	else
	    h = gethostbyname(arg);
	if (debug)
	    fprintf(stderr, "%s: %s (h_errno %d, h_trace %#o, rs_errno %d)\n",
		    progname, arg, h_errno, h_trace, rs_errno);
	if (h == NULL) {
	    fprintf(stderr, "%s: %s not found\n", progname, arg);
	    return(h_errno);
	}
	cc = res_mkquery(QUERY, h->h_name, C_IN, T_HINFO, (char *)0, 0,
			NULL, packet, sizeof(packet));
	if (cc < 0) {
		fprintf(stderr, "%s: res_mkquery: buffer too small\n",
			progname);
		return(-1);
	}
	cc = res_send(packet, cc, answer, sizeof(answer));
	cpu = os = "";
	hp = (HEADER *)answer;
	eom = answer + cc;
	if (cc >= 0 && hp->rcode == NOERROR && hp->ancount > 0) {
		cp = answer + sizeof(HEADER);
		if (cc = ntohs(hp->qdcount))
			while (--cc >= 0)
				cp += dn_skipname(cp, eom) + QFIXEDSZ;
		cp += dn_skipname(cp, eom) + RRFIXEDSZ;
		if (cc = *cp++) {
			cpu = cp;
			cp += cc;
		}
		if (cc = *cp) {
			*cp++ = '\0';
			os = cp;
			cp += cc;
			*cp = '\0';
		}
	}
	printf("%s", h->h_name);
	for (pp = h->h_aliases; *pp != NULL; pp++)
	    printf(" %s", *pp);
	first = 0;
	for (pp = h->h_addr_list; *pp != NULL; pp++) {
	    printf("%s%s", first++ ? ";" : ":",
		    inet_ntoa(*((struct in_addr *)(*pp))));
	    printf("%s%s%s%s", *cpu ? " CPU=" : "", cpu,
		    *os ? " OS=" : "", os);
	}
	printf("\n");
	break;
    case 'N':
	if (isdigit(*arg) && (network = inet_network(arg)) != -1) {
	    if ((n = getnetbyaddr(network, AF_INET)) == NULL) {
		fprintf(stderr, "%s: %s not found\n", progname, arg);
		return(-1);
	    }
	} else if ((n = getnetbyname(arg)) == NULL) {
	    fprintf(stderr, "%s: %s not found\n", progname, arg);
	    return(-1);
	}
	printf("%s", n->n_name);
	for (pp = n->n_aliases; *pp != NULL; pp++)
	    printf(" %s", *pp);
        (*(struct in_addr *)&addr) = inet_makeaddr(n->n_net, 0);
	printf(":%s\n", inet_nntoa(addr));
	break;
    case 'S':
	if ((s = getservbyname(arg, "tcp")) == NULL &&
	    (s = getservbyname(arg, "udp")) == NULL) {
	    fprintf(stderr, "%s: %s not found\n", progname, arg);
	    return(-1);
	}
	printf("%s", s->s_name);
	for (pp = s->s_aliases; *pp != NULL; pp++)
	    printf(" %s", *pp);
	printf(":%d/%s\n", ntohs((u_short)s->s_port), s->s_proto);
	break;
    case 'P':
	if ((p = getprotobyname(arg)) == NULL) {
	    fprintf(stderr, "%s: %s not found\n", progname, arg);
	    return(-1);
	}
	printf("%s", p->p_name);
	for (pp = p->p_aliases; *pp != NULL; pp++)
	    printf(" %s", *pp);
	printf(":%d\n", p->p_proto);
	break;
    }
    return(0);
}
