 /*
 **********************************************************************
 * HISTORY
 * $Log:	hostq.c,v $
 * Revision 2.7  91/08/21  14:22:16  parker
 * 	Added -cname option.  This option changes the default
 * 	behaviour, but makes it behave more like the "host" program.
 * 	It also forces consistant behaviour with regard to program output.
 * 	The first time a host is looked up (or before a CNAME record is
 * 	cached), it will print the CNAME returned.  Following lookups
 * 	(or after CNAME record is cached) only output canonical name. 
 * 	Also allows aliases to be printed when using host table.
 * 	[91/08/21            parker]
 * 
 * Revision 2.6  90/11/19  11:59:18  ern
 * 	Fix bug in compatibilitiy mode.
 * 	In compatibility mode instead of " not found", print real host error.
 * 	Fix major bug in handling of hosts without addreses.
 * 	[90/11/19  11:58:58  ern]
 * 
 * Revision 2.5  90/11/16  14:05:43  ern
 * 	For compatibility reasons with old hostq functionality add 
 * 	weird or inconsistent -noname switch.
 * 	[90/11/16  14:05:15  ern]
 * 
 * Revision 2.4  90/11/15  15:34:23  ern
 * 	Add -oneline switch. 
 * 	Add -name switch.
 * 	Make a few fixes to "host" compatiblity code.
 * 	[90/11/15  15:34:08  ern]
 * 
 * Revision 2.3  90/11/14  16:22:37  ern
 * 	no changes.
 * 	[90/11/14  16:22:02  ern]
 * 
 * 	Fixup exit codes to match existing host program.
 * 	Fixup so that when run as "host", it acts like existing host program.
 * 	Use a global variable for progrname and don't trash it.
 * 	When we do an gethostbyaddr() lookup, follow it up by a gethostbyname()
 * 	   to make sure we get the official name and not an alias. This should
 * 	   probably be done by the resolver.
 * 	Make sure we only process resource records that we care about not
 * 	   random ones that can give us garbage and/or make us die.
 * 	[90/11/14  16:11:44  ern]
 * 
 * Revision 2.2  88/09/29  17:33:04  mja
 * 	Change to correct for apparent _res_qname bug and to always
 * 	copy the host name into a local buffer to avoid clobbering of
 * 	the static data in subsequent queries.
 * 	[88/09/29  17:28:19  mja]
 * 	
 * 	Updated for the 4.3 nameserver domain.  Added request for
 * 	     MX records.  Added command line switches to request only 
 * 	specific information, in addition to the host name:
 * 		-addr	address
 * 		-os	operating system
 * 		-cpu	cpu type
 * 		-mx	MX records
 * 		-serv	well known services
 * 		-nolabl suppress labeling the output
 * 	
 * 	The default action is to return all information.  The host name is 
 * 	always printed.
 * 	
 * 	RENAMED to hostq to avoid breaking shell scripts.
 * 	
 * 	"Service" now accepts service name or number.
 * 	[88/08/29            raiff]
 * 
 * 20-Mar-86  Glenn Marcy (gm0w) at Carnegie-Mellon University
 *	Created.
 *
 **********************************************************************
 */
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <libc.h>
#include <c.h>

int debug = FALSE;
int eall, e1line, eaddr, eos, ecpu, emx, ewks, enolabl, ename, enoname, ecname;
int ehostmode;

char *prog = "host";

extern int h_trace;
extern int h_errno;
extern int rs_errno;

char **parseargs();
void switcherr();

/* 
 *   Valid option table.
 */
char *opts[] = {
    "-debug",
    "-oneline",
    "-name",
    "-noname",
    "-address",
    "-os",
    "-cpu",
    "-mx",
    "-service",
    "-nolabel",
    "-cname",
    0
};

/*
 *Valid opton types.  
 */

enum opt_types
{
    OPT_AMBIGUOUS=(-2),
    OPT_UNKNOWN,
    OPT_DEBUG,
    OPT_ONELINE,
    OPT_NAME,
    OPT_NONAME,
    OPT_ADDR,
    OPT_OS,
    OPT_CPU,
    OPT_MX,
    OPT_SERV,
    OPT_NOLABL,
    OPT_CNAME,
};



usage()
{
    char *tmp;

    tmp = prog;
    if (strcasecmp(prog,"hostq") == 0)
	tmp = "host";

    quit(-1, "usage: %s [ %s-name || %s-number ] ...\n", prog, tmp, tmp);
}

/* network number to ascii */
char *inet_nntoa(addr)
long addr;
{
    char *net;
    struct in_addr in;

    in.s_addr = addr;
    in.s_addr = ntohl((u_long)inet_netof(in));
    net = inet_ntoa(in);
    addr = ntohl((u_long)addr);
    if (IN_CLASSA(addr)) net += 6;
    else if (IN_CLASSB(addr)) net += 4;
    else if (IN_CLASSC(addr)) net += 2;
    return(net);
}

main(argc, argv)
int argc;
char **argv;
{
    char *prefix;
    int status = 0;

    ehostmode = FALSE;
    enoname = FALSE;
    eall = ename = e1line  = eaddr = eos = ecpu = emx = ewks = enolabl = ecname = FALSE;
    _res.options &= ~RES_DEFNAMES;

    prog = rindex(argv[0], '/');
    if (prog++ == NULL)
	prog = argv[0];
    argv++;

    if (strcasecmp(prog, "host") == 0) {
	prefix = "H";
	/* compatibilitiy with old "host" program */
	while (argc > 1 && **argv == '-') {
	    if (strcmp(*argv, "-d") == 0) {
		_res.options |= RES_DEBUG;
		debug++;
	    } else
		quit(-1, "%s: unknown switch '%s'\n", prog, *argv);
	    argc--;
	    argv++;
	}
	ehostmode = TRUE;
	enoname = FALSE;
	e1line = ename = eaddr = ecpu = eos = enolabl = TRUE;
    } else if (strcasecmp(prog, "hostq") == 0) {
	prefix = "H";
	argv = parseargs(argv);
	eall = !(ename | eaddr | ewks | eos | ecpu | emx);
    } else if (strcasecmp(prog, "net") == 0) {
	prefix = "N";
    } else if (strcasecmp(prog, "service") == 0) {
	if (argc < 2)
	    usage();
	prefix = "S";
    } else if (strncasecmp(prog, "protocol") == 0) {
	if (argc < 2)
	    usage();
	prefix = "P";
    } else
	quit(-1, "%s: must be called host, hostq, net, service or protocol\n", 
	     prog);

    if (argc == 1) {
	char *p, hostname[BUFSIZ];

	if (gethostname(hostname, sizeof(hostname)-2) < 0 ||
	    hostname[0] == '\0')
	    quit(-1, "couldn't find current hostname\n");
	p = hostname + strlen(hostname);
	if (p[-1] != '.') {
	    *p++ = '.';
	    *p = '\0';
	}
	if (*prefix == 'N') {
	    struct hostent *hp;

	    if ((hp = gethostbyname(hostname)) == NULL)
		quit(-1, "couldn't lookup host name %s\n", hostname);
	    (void)strcpy(hostname, inet_nntoa(*(long *)hp->h_addr));
	}
	exit(printent(prefix, hostname));
    }

    while (*argv) {
	int retval = printent(prefix, *argv);
	if (retval)
		status = retval;
	argv++;
    }
    exit(status);
}

static char *get_rr(reqtyp,cpp,eom,lenp)
int reqtyp;
char **cpp,*eom;
int *lenp;
{
	long thetype,theclass,thettl,thelen;
	char *cp;

	cp = *cpp;

	/* get resource record */
	cp += dn_skipname(cp, eom);
	thetype = _getshort(cp);
	cp += sizeof(u_short);
	theclass = _getshort(cp);
	cp += sizeof(u_short);
	thettl = _getlong(cp);
	cp += sizeof(u_long);
	thelen = _getshort(cp);
	cp += sizeof(u_short);

	/* update pointers */
	*lenp = thelen;
	*cpp = (cp + thelen);

	/* check type */
	if (thetype != reqtyp) 
	    return NULL;

	return cp;
}

static char *skip_q(cnt,cpp,eom)
int cnt;
char **cpp,*eom;
{
    char *cp;

    if (cnt <= 0)
	return;
    cp = *cpp;
    while (--cnt >= 0)
	cp += dn_skipname(cp, eom) + QFIXEDSZ;
    *cpp = cp;
}

mystrncpy(dest,src,len)
char *dest,*src;
int len;
{
    *dest = '\0';
    strncat(dest,src,len);
}

doprefix(str,force,nospace)
char *str;
int force,nospace;
{
    if (e1line && !nospace)
	printf(" ");

    if (!enolabl || force) {
	if (e1line)
	    printf("%s=",str);
	else
	    printf("%s: ",str);
    }
}

printent(prefix, arg)
char *prefix, *arg;
{
    struct hostent *h;
    struct netent *n;
    struct servent *s;
    struct protoent *p;
    char *cpu, *os;
    char *cp, **pp, *cp1;
    int first, cc;
    struct in_addr inaddr, inet_makeaddr();
    long addr;
    int network;
    char packet[PACKETSZ];
    char answer[PACKETSZ];
    HEADER *hp;
    char name[MAXDNAME];
    int i, j, tmp;
    int pref, dlen;
    char *eom, hostname[MAXDNAME+1], *previous, *thisaddr;
    extern char _res_qname[];
    char tmpbuf1[BUFSIZ],tmpbuf2[BUFSIZ];
    int exit_code = 0;


    switch (*prefix) {
    case 'H':
	if (isdigit(*arg) && (addr = inet_addr(arg)) != -1) {
	    h = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);

	    /* if we got a name from an address then reverse...gets around
	     * following cnames even though the resolve should really do
	     * this for us.
	     */
	    if (h != NULL) {
		strcpy(name,h->h_name);
		strcat(name,".");	/* add trailing dot */
		h = gethostbyname(name);
	    }
	}
	else
	    h = gethostbyname(arg);
	if (debug)
	    fprintf(stderr, "%s: %s (h_errno %d, h_trace %#o, rs_errno %d)\n",
		    prog, arg, h_errno, h_trace, rs_errno);

	if (h == NULL && (h_errno != NO_ADDRESS || ehostmode)) {
	    sprintf(tmpbuf1,"%s: %s",prog, arg);
	    herror(tmpbuf1);
	    return (h_errno);
	}

	strcpy(hostname, (h == NULL)
			 ?
			   (_res_qname[0] != '\0' ? _res_qname : arg)
			 :
			   h->h_name);
	if (eall || ename || !enoname){
	    doprefix("Name",0,1);
	    printf("%s", hostname);
	    if ((h != NULL) && ecname)
		for (pp = h->h_aliases; *pp != NULL; pp++)
		    printf(" %s", *pp);
	    if (!e1line)
		printf("\n");
	}

	if (debug) printf("ADDRESS LIST:\n");

	if (h != NULL) {
	    if (eall || eaddr){
	        first = 0;
		if (eaddr && !*h->h_addr_list) exit_code = 2;
		if (!ehostmode) {
		    doprefix("Address(es)",0,0);
		    for (pp = h->h_addr_list; *pp != NULL; pp++) {
			if(first++) printf(" ");
			printf("%s",inet_ntoa(*((struct in_addr *)(*pp))));
		    }
		    if(!e1line && first)
			printf("\n");
		}
	    }
	}

	if (debug) printf("CPU/OS LIST:\n");

	if (eall || eos || ecpu || ehostmode){
	    cc = res_mkquery(QUERY, hostname, C_IN, T_HINFO, (char *)0, 0,
		 	    NULL, packet, sizeof(packet));
	    if (cc < 0) {
		fprintf(stderr, "%s: res_mkquery: buffer too small\n",
			prog);
		return(-1);
	    }
	    cc = res_send(packet, cc, answer, sizeof(answer));
	    cpu = os = "";
	    hp = (HEADER *)answer;
	    if (cc >= 0 && hp->rcode == NOERROR && hp->ancount != 0) {
		cp = answer + sizeof(HEADER);
		eom = (char *)answer + cc;
		skip_q(ntohs(hp->qdcount),&cp,eom);
		/* process RR records which are variable size */
		if (cc = ntohs(hp->ancount)) {
		    while (--cc >= 0) {
			cp1 = get_rr(T_HINFO,&cp,eom,&dlen);
			if (cp1 == NULL)
			    continue;

			/* got what we want, break it out */
			if (cc = *cp1++) {
			    mystrncpy(cpu = tmpbuf1,cp1,cc);
			    cp1 += cc;
			}
			if (cc = *cp1++) {
			    mystrncpy(os = tmpbuf2,cp1,cc);
			    cp1 += cc;
			}
			break;
		    }
		}
	    }
	    first = 0;
	    if (h == NULL)
		pp = NULL;
	    else
		pp = h->h_addr_list;
	    for(;;) {
		if (ehostmode && pp != NULL && *pp != NULL) {
		    printf("%s",first++ ? ";" : ":");
		    printf("%s",inet_ntoa(*((struct in_addr *)(*pp))));
		}
		if (eall || ecpu) {
		    if (*cpu) {
			doprefix("CPU",ehostmode,0);
			printf("%s", cpu);
			if (!e1line)
			    printf("\n");
		    }
		    else {
			if (ecpu) exit_code = 2;
		    }
		}
		if (eall || eos) {
		    if (*os) {
			doprefix("OS",ehostmode,0);
			printf("%s", os);
			if (!e1line)
			    printf("\n");
		    }
		    else {
			if (eos) exit_code = 2;
		    }
		}
		/* if no address or hostmode stop */
		if (pp == NULL || *pp++ == NULL || !ehostmode)
		    break;
		/* or no more addresses */
		if (*pp == NULL)
		    break;
	    }
	}

	if (debug) printf("MX LIST:\n");

	if (eall || emx){
	    cc = res_mkquery(QUERY, hostname, C_IN, T_MX, (char *)0, 0,
			    NULL, packet, sizeof(packet));
	    if (cc < 0) {
		fprintf(stderr, "%s: res_mkquery: buffer too small\n",
			prog);
		return (-1);
	    }
	    cc = res_send(packet, cc, answer, sizeof(answer));
	    hp = (HEADER *)answer;
	    if (cc >= 0 && hp->rcode == NOERROR && hp->ancount != 0) {
		cp = answer + sizeof(HEADER);
		eom = (char *)answer + cc;
		skip_q(ntohs(hp->qdcount),&cp,eom);
		/* process RR records which are variable size */
		if (cc = ntohs(hp->ancount)) {
		    first = 1;
		    while (--cc >= 0) {
			cp1 = get_rr(T_MX,&cp,eom,&dlen);
			if (cp1 == NULL)
			    continue;

			if (first)
			    doprefix("MX",0,0);

			/* got what we want, break it out */
			pref = _getshort(cp1);
			cp1 += sizeof(u_short);
			tmp = dn_expand(answer, eom, cp1,
				name, sizeof(name));
			if (name[0] == '\0'){
			    name[0] = '.';
			    name[1] = '\0';
			}
			printf("%s;%ld ",name ,pref);
			first = 0;
		    }
		    if (!first && !e1line)
			printf("\n");
		}
	    } else {
		if (emx) exit_code = 2;
	    }
	}

	if (debug) printf("WKS LIST:\n");

	if (eall || ewks){
	    cc = res_mkquery(QUERY, hostname, C_IN, T_WKS, (char *)0, 0,
			    NULL, packet, sizeof(packet));
	    if (cc < 0) {
		fprintf(stderr, "%s: res_mkquery: buffer too small\n",
			prog);
		return(-1);
	    }
	    cc = res_send(packet, cc, answer, sizeof(answer));
	    hp = (HEADER *)answer;
	    if ( cc >= 0 && hp->rcode == NOERROR && hp->ancount != 0) {
		cp = answer + sizeof(HEADER);
		eom = (char *)answer + cc;
		skip_q(ntohs(hp->qdcount),&cp,eom);
		/* process RR records which are variable size */
		if (cc = ntohs(hp->ancount)) {
		    first = 1;
		    while (--cc >= 0) {
			cp1 = get_rr(T_WKS,&cp,eom,&dlen);
			if (cp1 == NULL)
			    continue;

			if (dlen <= sizeof(u_long))
			    continue;

			/* get ip address */
			bcopy(cp1, (char *)&inaddr, sizeof(inaddr));
			cp1 += sizeof(u_long);
			dlen -= sizeof(u_long);
			thisaddr = inet_ntoa(inaddr);
			if (first || strcasecmp(tmpbuf1, thisaddr) != 0){
			    char tmpbuf[128];
			    sprintf(tmpbuf,"Services for %s",thisaddr);
			    doprefix(tmpbuf,0,0);
			    first = 0;
			}
			p = getprotobynumber(*cp1++);
			i = 0;
			while (dlen-- > 0) {
			    tmp = *cp1++ & 0xff;
			    if (p != NULL) do {
				if (tmp & 0200) {
				    s = getservbyport(htons(i),p->p_name);
				    if (s != NULL)
					printf("%s/%s ",p->p_name,s->s_name);
				}
				tmp <<= 1;
			    } while ((++i & 07) != 0);
			}
			strcpy(tmpbuf1, thisaddr);
		    }
		    if (!e1line && !first)
			printf("\n");
		}
	    } else {
		if (ewks) exit_code = 2;
	    }
	}
	if (e1line)
	    printf("\n");
	break;
    case 'N':
	if (isdigit(*arg) && (network = inet_network(arg)) != -1) {
	    if ((n = getnetbyaddr(network, AF_INET)) == NULL) {
		fprintf(stderr, "%s: %s not found\n", prog, arg);
		return(-1);
	    }
	} else if ((n = getnetbyname(arg)) == NULL) {
	    fprintf(stderr, "%s: %s not found\n", prog, arg);
	    return(-1);
	}
	printf("%s", n->n_name);
	for (pp = n->n_aliases; *pp != NULL; pp++)
	    printf(" %s", *pp);
        (*(struct in_addr *)&addr) = inet_makeaddr(n->n_net, 0);
	printf(":%s\n", inet_nntoa(addr));
	break;
    case 'S':
    	if (isdigit(*arg)) {
		if((s = getservbyport(htons(atoi(arg)), "tcp")) == NULL) {
			s = getservbyport(htons(atoi(arg)), "udp");
		}
	} else if ((s = getservbyname(arg, "tcp")) == NULL)
	    s = getservbyname(arg, "udp");
	if (s == NULL) {
	    fprintf(stderr, "%s: %s not found\n", prog, arg);
	    return(-1);
	}
	printf("%s", s->s_name);
	for (pp = s->s_aliases; *pp != NULL; pp++)
	    printf(" %s", *pp);
	printf(":%d/%s\n", ntohs((u_short)s->s_port), s->s_proto);
	break;
    case 'P':
	if ((p = getprotobyname(arg)) == NULL) {
	    fprintf(stderr, "%s: %s not found\n", prog, arg);
	    return(-1);
	}
	printf("%s", p->p_name);
	for (pp = p->p_aliases; *pp != NULL; pp++)
	    printf(" %s", *pp);
	printf(":%d\n", p->p_proto);
	break;
    }
    return(exit_code);
}



char **parseargs(argv)
char **argv;
{
    int opt;
    
    for (; *argv; argv++) {
	if (*argv[0] != '-')
	    break;
	switch(opt = stablk(*argv, opts, TRUE))
	{
	    case OPT_AMBIGUOUS:
	    {
		quit(-2, "%s: ambiguous option %s\n", prog, *argv);
		break;
	    }
	    case OPT_UNKNOWN:
	    {
				switcherr();
				break;
	    }
	    case OPT_DEBUG:	_res.options |= RES_DEBUG;
	    			debug++;
				break;
	    case OPT_NAME:	ename++;
	    			break;
	    case OPT_NONAME:	enoname++;
	    			break;
	    case OPT_ONELINE:	e1line++;
	    			break;
	    case OPT_ADDR:	eaddr++;
	    			break;
	    case OPT_OS:	eos++;
				break;
	    case OPT_CPU:	ecpu++;
	    			break;
	    case OPT_MX:	emx++;
				break;
	    case OPT_SERV:	ewks++;
	    			break;
	    case OPT_NOLABL:	enolabl++;
				break;
	    case OPT_CNAME:	ecname++;
				break;

	}
    }
    return(argv);
}
	    

void switcherr()
{
    fprintf(stderr, "Usage: %s [option] [machine]\n", prog);
    fprintf(stderr, "\nValid options are:\n");
    fprintf(stderr, "\t-name \tprint machine's name\n");
    fprintf(stderr, "\t-cname \tprint machine's cname(s)\n");
    fprintf(stderr, "\t-address \tprint machine's address\n");
    fprintf(stderr, "\t-os \t\tprint machine's operationg system\n");
    fprintf(stderr, "\t-cpu \t\tprint machine's cpu type\n");
    fprintf(stderr, "\t-mx \t\tprint mx records\n");
    fprintf(stderr, "\t-service \tprint services supported by machine\n");
    fprintf(stderr, "\t-nolabel \tsupress printing of field names\n");
    fprintf(stderr, "\t-oneline \tsupress printing of multiple newlines\n");
    quit(-2);
}    
