/*
 * Copyright (c) 1991 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator   or   Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the rights
 * to redistribute these changes.
 *
 *
 * whenis: program which parses a date and prints it out in a nice
 * format. Default prints out the current date and time.
 * 
 * whenis [options] [date-time]
 * 
 * -p: Assumes past date instead of future.
 *
 * -b: Brief form of -c and pretty (default) output.
 * -c: Prints date in ctime(3) format.
 * -e: Make unrecognizable tokens terminate the parse.
 * -f: Prints date according to following format specification.
 * -n: Prints days since 1/1/1900 and seconds after midnight.
 * -s: Prints first eight fields of tm structure in the format:
 *     yyyy mm/dd hh:mm:ss ddd w
 *     (year month/day time year-day week-day)
 * -x: Execute resultant date as a command.
 * -N: The next argument is a numeric time which is printed in the
 *     selected format.
 * 
 * Copyright (c) 1984, Leonard G C Hamey.
 * 
 * HISTORY
 * $Log:	whenis.c,v $
 * Revision 1.3  91/02/18  11:28:51  gln
 * 	Added the CMU copyright disclaimer.
 * 	[91/02/18  11:28:21  gln]
 * 
 * Revision 1.2  88/11/16  22:07:17  gm0w
 * 	Added -L, -A, -M, and -C switches for using file times for
 * 	the reference date.
 * 	[88/11/16            gm0w]
 * 
 * 7-Mar-84  Leonard Hamey (lgh) at Carnegie-Mellon University
 *    Added -e option.
 * 13-Feb-84 Leonard Hamey (lgh) at Carnegie-Mellon University
 *    Added -f option and -x option
 * 6-Feb-84  Leonard Hamey (lgh) at Carnegie-Mellon University
 *    Added -N option
 * 3-Feb-84  Leonard Hamey (lgh) at Carnegie-Mellon University
 *    Added -b option.
 * 23-Jan-84 Leonard Hamey (lgh) at Carnegie-Mellon University
 *    Added -p option for past dates.
 * 17-Jan-84 Leonard Hamey (lgh) at Carnegie-Mellon University
 *    Created.
 */

# include <sys/types.h>
# include <sys/stat.h>
# include <sys/time.h>

main (ac, av)
int ac;
char **av;
{ char argdate[100];
  struct stat st;
  struct tm *mytime, *localtime (), mytm;
  long time (), gtime (), ntime, atol ();
  char *asctime (), *asciitime (), *fdate ();
  int optb = 0, optc = 0, optn = 0, opts = 0, now = 0, optp = 0;
  int optN = 0, optf = 0, optx = 0, opte = 0;
  int optL = 0;
  char *form;
  char text[100];
  char c;
  char *textp, *argv[50], *nxtarg ();
  int i;
  mytime = &mytm;
  while (ac > 1 && *av[1] == '-')
  { c = av[1][1];
    while (1)
    { if (c == 'b')
	optb++;
      else if (c == 'c')
	optc++;
      else if (c == 'x')
	optx++;
      else if (c == 'e')
	opte++;
      else if (c == 'n')
	optn++;
      else if (c == 'p')
	optp++;
      else if (c == 's')
	opts++;
      else if (c == 'L')
	optL++;
      else
	break;
      av[1]++;
      c = av[1][1];
    }
    if (av[1][1] == 'f')
    { optf++;
      ac--;
      av++;
      if (ac < 1)
	quit (1, "Option -f requires following format.\n");
      form = av[1];
    }
    else if (av[1][1] == 'N')
    { optN++;
      ac--;
      av++;
      if (ac < 1)
	quit (1, "Option -N requires following numeric time.\n");
      ntime = atol (av[1]);
    }
    else if (av[1][1] == 'A')
    { optN++;
      ac--;
      av++;
      if (ac < 1)
	quit (1, "Option -A requires following filename.\n");
      if ((optL?lstat(av[1], &st):stat(av[1], &st)) < 0)
	quit (1, "File %s not found.\n", av[1]);
      ntime = st.st_atime;
    }
    else if (av[1][1] == 'M')
    { optN++;
      ac--;
      av++;
      if (ac < 1)
	quit (1, "Option -M requires following filename.\n");
      if ((optL?lstat(av[1], &st):stat(av[1], &st)) < 0)
	quit (1, "File %s not found.\n", av[1]);
      ntime = st.st_mtime;
    }
    else if (av[1][1] == 'C')
    { optN++;
      ac--;
      av++;
      if (ac < 1)
	quit (1, "Option -C requires following filename.\n");
      if ((optL?lstat(av[1], &st):stat(av[1], &st)) < 0)
	quit (1, "File %s not found.\n", av[1]);
      ntime = st.st_ctime;
    }
    else if (av[1][1] != '\0')
      quit (1,
        "Usage: whenis [-bcenpsx] [-f format] [-N num-time] [date-time]\n");
    ac--;
    av++;
  }
  if (! optN)
  { ntime = time (0);
  }
  *mytime = *localtime (&ntime);
  if (ac > 1)
  { argdate[0] = 0;
    for (i = 1; i < ac; i++)
    { strcat (argdate, av[i]);
      strcat (argdate, " ");
    }
    if (parsedate (argdate, mytime, 0, optp, opte == 0) != 0)
    { printf ("I don't understand %s\n", argdate);
      exit (-1);
    }
  }
  else
  { now = ! optN;
  }
  if (optc)
  { if (optb)
      printf ("%.16s\n", asctime (mytime));
    else
      printf ("%s", asctime (mytime));
  }
  else if (optn)
  { if (mytime->tm_hour == -1)
    { mytime->tm_hour = 23;
      mytime->tm_min = 59;
      mytime->tm_sec = 0;
    }
    printf ("%ld\n", gtime (mytime));
  }
  else if (opts)
    printf ("%04d %02d/%02d %02d:%02d:%02d %3d %d\n", mytime->tm_year + 1900,
      mytime->tm_mon + 1, mytime->tm_mday, mytime->tm_hour, mytime->tm_min,
      mytime->tm_sec, mytime->tm_yday, mytime->tm_wday);
  else if (optf)
  { if (now)
      mytime->tm_sec = 0;
    if (optx)
    { textp = fdate (text, form, mytime);
      i = 0;
      while (*textp)
      { argv[i] = nxtarg (&textp, " \t");
	i++;
      }
      argv[i] = 0;
      execvp (argv[0], argv);
      quit (1, "Exec of %s failed.\n", argv[0]);
    }
    printf ("%s\n", fdate (text, form, mytime));
  }
  else
  { if (now)
      mytime->tm_sec = 0; 		     /* dont show seconds */
/*    printf ("%s\n", asciitime (mytime, optb)); */
    if (optb)
      printf ("%s\n", fdate (text, "%3Wday %3Month %day{,{ %?year}{ %time}}",
        mytime));
    else
      printf ("%s\n", fdate (text,
        "%Wday %Month %day%th, %year{ [%midnight|%noon|%0hour:%min:%?sec|{%hour:%min %am}]",
        mytime));
  }
}

static char *days[] =
{ "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"
};

static char *months[] =
{ "January", "February", "March", "April", "May", "June",
  "July", "August", "September", "October", "November", "December"
};

static char result[50];

char *asciitime (tm, brief)
struct tm *tm;
int brief;
{ long ctime, time ();
  struct tm ctm, *localtime ();
  register int ddiff;
  char nstr[20];
  if (brief)
  { ctime = time (0);
    ctm = *localtime (&ctime);
  }
  result[0] = '\0';
  if (tm->tm_wday >= 0)
  { if (brief)
      strncat (result, days[tm->tm_wday], 3);
    else
      strcat (result, days[tm->tm_wday]);
    strcat (result, " ");
  }
  if (tm->tm_mon >= 0)
  { if (brief)
      strncat (result, months[tm->tm_mon], 3);
    else
      strcat (result, months[tm->tm_mon]);
    strcat (result, " ");
  }
  if (tm->tm_mday >= 0)
  { sprintf (nstr, "%d", tm->tm_mday);
    strcat (result, nstr);
    if (! brief)
    { strcat (result, (tm->tm_mday / 10) % 10 == 1 ? "th" :
        tm->tm_mday % 10 == 1 ? "st" :
        tm->tm_mday % 10 == 2 ? "nd" :
        tm->tm_mday % 10 == 3 ? "rd" :
        "th");
    }
  }
  ddiff = (tm->tm_year - ctm.tm_year) * 366 + tm->tm_yday - ctm.tm_yday;
  if (! brief || ddiff < 0 || ddiff > 300)
  { sprintf (nstr, ", %d", tm->tm_year + 1900);
    strcat (result, nstr);
  }
  else if (tm->tm_hour >= 0)
    strcat (result, ",");
  if (tm->tm_hour >= 0)
  { strcat (result, " ");
    if (tm->tm_sec > 0)
    { sprintf (nstr, "%02d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->tm_sec);
      strcat (result, nstr);
    }
    else if (tm->tm_hour == 0 && tm->tm_min == 0)
    { strcat (result, "00:00");
    }
    else if (tm->tm_hour == 12 && tm->tm_min == 0)
      strcat (result, "noon");
    else
    { sprintf (nstr, "%d", (tm->tm_hour + 11) % 12 + 1);
      strcat (result, nstr);
      if (! brief || tm->tm_min != 0)
      { sprintf (nstr, ":%02d", tm->tm_min);
	strcat (result, nstr);
      }
      if (! brief)
	strcat (result, " ");
      strcat (result, tm->tm_hour >= 12 ? "pm" : "am");
    }
  }
  return (result);
}
