/*
 * winsort.c   WINdow SORT
 * 
 *     Only sorts over a window.  If a file is nearly sorted already then 
 * this algorithm will run very fast and use very little memory.  If the 
 * file is not "near" enough to sorted the program will terminate with 
 * an error message.
 *
 *     The program maintains a queue of lines sorted by an integer number in
 * the first field.  A new element is added to the queue after the first
 * element that is smaller than it.  If the queue is filling up memory
 * then a few elements off the front of the queue are output.  If an element 
 * is ever smaller than everything in the queue and some output has already 
 * been written then the file was not "near" enough to sorted and an the 
 * program terminates.
 *
 * Head has a key of 0,
 * tail has a key of maxint.
 */

#include <stdio.h>

#define MaxElements 10000       /* 10000 lines * 80 bytes/line = 800 KB */
#define FALSE 0
#define TRUE  1

int NothingOut;
int Elements;

typedef struct line {
		         double key;	
   			     char data[100];
		struct   line *next;
		struct   line *prev;
	} LINE;

LINE *head, *tail;



InsertLine(newline)
LINE *newline;
{
	LINE *ptr, *next;
	int status;

	status = sscanf(newline->data,"%lf",&newline->key);
	if (status == EOF) {
		fprintf(stderr, "ERROR: did not find an integer on the line.\n");
		exit(1);
	}

	for (ptr=tail; ptr->key > newline->key; ptr=ptr->prev) /* works for key=0 */
		;
	if ((ptr==head) && (NothingOut==FALSE)) {
		fprintf(stderr,"Sort window was not big enough for this file.\n");
		fprintf(stderr,"Increase the MaxElements constant.\n");
		fprintf(stderr,"%s",head->next->data);
		fprintf(stderr,"%s",newline->data);
		fprintf(stderr,"%s",tail->prev->data);
		exit(1);	
	}
	next = ptr->next;
	newline->next=next;
	newline->prev=ptr;
	next->prev=newline;
	ptr->next=newline;
	Elements++;
}

int Output1()
{
LINE *out, *outnext;

 	out=head->next;
	if (out==tail)
		return(FALSE);
	printf("%s",out->data);

	outnext=out->next;
	head->next=outnext;
	outnext->prev=head;
	free(out);
	Elements--;
	NothingOut=FALSE;
	return(TRUE);
}

OutputAll()
{
	while(Output1())
		;
}

init()
{
	head = (LINE * ) malloc(sizeof(LINE));
	tail = (LINE * ) malloc(sizeof(LINE));

	head->prev=NULL;
	head->next=tail;
	head->key= -1e20;

	tail->prev=head;
	tail->next=NULL;
	tail->key= 1e20;     /* everything should be between head and tail */

	Elements=0;
	NothingOut=TRUE;
}

main()
{
	LINE *newline;
	char *tmp;

    init();
	while ( !feof(stdin) ) {
		newline = (LINE *) malloc(sizeof(LINE));
		tmp=fgets(newline->data,100,stdin);
		if (tmp==NULL)                         
			break;
		InsertLine(newline);
		if (Elements > MaxElements) {
			Output1();
		}
	}
	OutputAll();
	if ( !feof(stdin) ) 
		fprintf(stderr,"Input error from fgets.\n");
}


